"""Identify and anonymize IP addresses."""
#   Copyright 2018 Intentionet
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from abc import ABCMeta, abstractmethod

from bidict import bidict
import ipaddress
import logging
# Need regex instead of re for variable look behind
import regex

from hashlib import md5
from six import add_metaclass, iteritems, text_type, u


_IPv4_OCTET_PATTERN = r'(25[0-5]|(2[0-4]|1?[0-9])?[0-9])'

# Deliberately allowing leading zeros and will remove them later
IPv4_PATTERN = regex.compile(
    r'(?<=^|[\s:<>/\'",=\(])((0*{octet}\.){{3}}0*{octet})'
    r'(?=/(\d{{1,3}}))?(?=[-\s:<>/\'",=\]\)]|$)'.format(octet=_IPv4_OCTET_PATTERN))

# Modified from https://stackoverflow.com/a/17871737/1715495
IPv6_PATTERN = regex.compile(
    r'(?<=^|[\s<>/\'",=\(])(([0-9a-f]{1,4}:){7,7}[0-9a-f]{1,4}'
    r'|([0-9a-f]{1,4}:){1,7}:'
    r'|([0-9a-f]{1,4}:){1,6}:[0-9a-f]{1,4}'
    r'|([0-9a-f]{1,4}:){1,5}(:[0-9a-f]{1,4}){1,2}'
    r'|([0-9a-f]{1,4}:){1,4}(:[0-9a-f]{1,4}){1,3}'
    r'|([0-9a-f]{1,4}:){1,3}(:[0-9a-f]{1,4}){1,4}'
    r'|([0-9a-f]{1,4}:){1,2}(:[0-9a-f]{1,4}){1,5}'
    r'|[0-9a-f]{1,4}:((:[0-9a-f]{1,4}){1,6})'
    r'|:((:[0-9a-f]{1,4}){1,7}|:)'
    r'|fe80:(:[0-9a-f]{0,4}){0,4}%[0-9a-z]{1,}' +
    r'|::(ffff(:0{{1,4}})?:)?({octet}\.){{3}}{octet}'
    r'|([0-9a-f]{{1,4}}:){{1,4}}:({octet}\.){{3}}{octet})'.format(
        octet=_IPv4_OCTET_PATTERN) +
    r'(?=[-\s<>/\'",=\]\)]|$)',
    regex.IGNORECASE)


def _generate_bit_from_hash(salt, string):
    """Return the last bit of the result from hashing the input string."""
    last_hash_digit = md5((salt + string).encode()).hexdigest()[-1]
    return int(last_hash_digit, 16) & 1


def _ensure_unicode(str):
    if not isinstance(str, text_type):
        str = u(str)
    return str


@add_metaclass(ABCMeta)
class _BaseIpAnonymizer(object):
    def __init__(self, salt, length, salter=_generate_bit_from_hash):
        self.salt = salt
        self.cache = bidict({'': ''})
        self.length = length
        self.fmt = '{{:0{length}b}}'.format(length=length)
        self.salter = salter

    def anonymize(self, ip_int):
        bits = self.fmt.format(ip_int)
        anon_bits = self._anonymize_bits(bits)
        return int(anon_bits, 2)

    def _anonymize_bits(self, bits):
        ret = self.cache.get(bits)
        if ret is not None:
            return ret

        head, last = bits[:-1], int(bits[-1])
        flip_last = self.salter(self.salt, head)
        ret = self._anonymize_bits(head) + str(flip_last ^ last)

        # Cache before returning.
        self.cache[bits] = ret
        return ret

    def deanonymize(self, ip_int):
        bits = self.fmt.format(ip_int)
        anon_bits = self._deanonymize_bits(bits)
        return int(anon_bits, 2)

    def _deanonymize_bits(self, bits):
        ret = self.cache.inv.get(bits)
        if ret is not None:
            return ret

        head, last = bits[:-1], int(bits[-1])
        orig_head = self._deanonymize_bits(head)
        flip_last = self.salter(self.salt, orig_head)
        ret = orig_head + str(flip_last ^ last)

        # Cache before returning.
        self.cache.inv[bits] = ret
        return ret

    def dump_to_file(self, file_out):
        ips = ((bits, anon_bits)
               for bits, anon_bits in iteritems(self.cache)
               if len(bits) == self.length)
        for bits, anon_bits in ips:
            ip = self._ip_to_str(bits)
            anon = self._ip_to_str(anon_bits)
            file_out.write('{}\t{}\n'.format(ip, anon))

    @classmethod
    def _ip_to_str(cls, bits):
        return str(cls.make_addr_from_int(int(bits, 2)))

    @classmethod
    @abstractmethod
    def get_addr_pattern(cls):
        raise NotImplementedError()

    @classmethod
    @abstractmethod
    def make_addr(cls, addr_str):
        raise NotImplementedError()

    @classmethod
    @abstractmethod
    def make_addr_from_int(cls, ip_int):
        raise NotImplementedError()

    @abstractmethod
    def should_anonymize(self, ip_int):
        raise NotImplementedError()


class IpAnonymizer(_BaseIpAnonymizer):
    """An anonymizer for IPv4 addresses."""

    _DROP_ZEROS_PATTERN = regex.compile(r'0*(\d+)\.0*(\d+)\.0*(\d+)\.0*(\d+)')

    def __init__(self, salt, **kwargs):
        """Create an anonymizer using the specified salt."""
        super(IpAnonymizer, self).__init__(salt, 32, **kwargs)
        # preserve IPv4 classes
        for i in range(4):
            bits = '1' * i
            self.cache[bits + '1'] = bits + '1'
            self.cache[bits + '0'] = bits + '0'

    def _is_mask(self, possible_mask_int):
        """Return True if the input int can be used as a 32-bit prefix mask.

        An IP address used as a prefix mask in IPv4 is either 1s followed by 0s,
        or the reverse. For example, 128.0.0.0, 0.0.63.255, etc.

        0.0.0.0 and 255.255.255.255 are considered masks.
        """
        # Implemented by counting the transitions between 1 and 0. All masks
        # will have at most one transition.
        diff = (possible_mask_int ^ (possible_mask_int >> 1)) & 0x7FFFFFFF
        # This code is a bit twiddle to determine if diff has at most 1 bit set.
        return (diff & ((0xFFFFFFFF ^ diff) + 1)) == diff

    @classmethod
    def get_addr_pattern(cls):
        """Return a compiled regex pattern to recognize IPv4 addresses."""
        return IPv4_PATTERN

    @classmethod
    def make_addr(cls, addr_str):
        """
        Return an IPv4 address from the given string.

        If the octets in `addr_str` have leading zeros, such as in 1.2.3.040,
        those zeros will be ignored (1.2.3.40) -- they will NOT be interpreted
        as octal (1.2.3.32).
        """
        addr_str = IpAnonymizer._DROP_ZEROS_PATTERN.sub(r'\1.\2.\3.\4', addr_str)
        return ipaddress.IPv4Address(_ensure_unicode(addr_str))

    @classmethod
    def make_addr_from_int(cls, ip_int):
        """Return an IPv4 address with the given int representation."""
        return ipaddress.IPv4Address(ip_int)

    def should_anonymize(self, ip_int):
        """Check if a given address should be anonymized (e.g. is it a mask or address?)."""
        return not self._is_mask(ip_int)


class IpV6Anonymizer(_BaseIpAnonymizer):
    """An anonymizer for IPv6 addresses."""

    def __init__(self, salt, **kwargs):
        """Create an anonymizer using the specified salt."""
        super(IpV6Anonymizer, self).__init__(salt, 128, **kwargs)

    @classmethod
    def get_addr_pattern(cls):
        """Return a compiled regex pattern to recognize IPv6 addresses."""
        return IPv6_PATTERN

    @classmethod
    def make_addr(cls, addr_str):
        """Return an IPv6 address from the given string."""
        return ipaddress.IPv6Address(_ensure_unicode(addr_str))

    @classmethod
    def make_addr_from_int(cls, ip_int):
        """Return an IPv6 address with the given int representation."""
        return ipaddress.IPv6Address(ip_int)

    def should_anonymize(self, ip_int):
        """Check if a given address should be anonymized."""
        return True


def _anonymize_match(anonymizer, match, undo_ip_anon):
    ip = anonymizer.make_addr(match)
    ip_int = int(ip)
    if not anonymizer.should_anonymize(ip_int):
        logging.debug("Should not anonymize %s, skipping", ip)
        return match

    if undo_ip_anon:
        new_ip_int = anonymizer.deanonymize(ip_int)
    else:
        new_ip_int = anonymizer.anonymize(ip_int)
    new_ip = anonymizer.make_addr_from_int(new_ip_int)
    logging.debug("Replacing %s with %s", ip, new_ip)
    return str(new_ip)


def anonymize_ip_addr(anonymizer, line, undo_ip_anon=False):
    """Replace each IP address in the line with an anonymized IP address.

    Masks will be unchanged. That is, any IP address that, when written in
    binary, that consists solely of an initial group of 1s followed by 0s
    or initial 0s followed by 1s will be unchanged.

    If undo_ip_anon is True, then each IP address encountered will be
    treated as an address already anonymized using the specified salt, and it
    will be replaced with the unanonymized address.
    """
    pattern = anonymizer.get_addr_pattern()
    return pattern.sub(lambda match: _anonymize_match(anonymizer, match[0], undo_ip_anon), line)



"""Default reserved words that should not be anonymized by sensitive word anonymization."""
#   Copyright 2018 Intentionet
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

# These are important tokens in network device configs and should not be modified by keyword anonymization
default_reserved_words = {
    '3des',
    '3des-cbc',
    '3des-sha1',
    '802.3ad',
    'aaa',
    'aaa-profile',
    'aaa-server',
    'aaa-user',
    'absolute-timeout',
    'accept',
    'accept-data',
    'accept-dialin',
    'accept-lifetime',
    'accept-own',
    'accept-register',
    'accept-rp',
    'accepted-prefix-limit',
    'access',
    'access-class',
    'access-group',
    'access-list',
    'access-log',
    'access-map',
    'access-profile',
    'accounting',
    'accounting-list',
    'accounting-port',
    'accounting-server-group',
    'accounting-threshold',
    'acct-port',
    'acfe',
    'ack',
    'acl',
    'acl-policy',
    'acllog',
    'action',
    'action-type',
    'activate',
    'activate-service-whitelist',
    'activation-character',
    'active',
    'active-server-group',
    'add',
    'add-path',
    'add-vlan',
    'additional-paths',
    'additive',
    'address',
    'address-book',
    'address-family',
    'address-hiding',
    'address-mask',
    'address-pool',
    'address-pools',
    'address-range',
    'address-set',
    'address-table',
    'addrgroup',
    'adjacency',
    'adjacency-check',
    'adjacency-stale-timer',
    'adjmgr',
    'admin',
    'admin-dist',
    'admin-distance',
    'admin-state',
    'admin-vdc',
    'administrative-weight',
    'admission',
    'admission-control',
    'adp',
    'advertise',
    'advertise-external',
    'advertise-inactive',
    'advertise-interval',
    'advertise-peer-as',
    'advertisement-interval',
    'aes',
    'aes-128-cbc',
    'aes-128-cmac-96',
    'aes-192-cbc',
    'aes-256-cbc',
    'aes128',
    'aes128-sha1',
    'aes192',
    'aes256',
    'aes256-sha1',
    'aesa',
    'af-group',
    'af11',
    'af12',
    'af13',
    'af21',
    'af22',
    'af23',
    'af31',
    'af32',
    'af33',
    'af41',
    'af42',
    'af43',
    'affinity',
    'affinity-map',
    'afs',
    'age',
    'aggregate',
    'aggregate-address',
    'aggregated-ether-options',
    'aggressive',
    'aging',
    'ah',
    'ahp',
    'airgroup',
    'airgroupservice',
    'ais-shut',
    'alarm',
    'alarm-report',
    'alert-group',
    'alerts',
    'alg',
    'alg-based-cac',
    'algorithm',
    'alias',
    'aliases',
    'all',
    'all-alarms',
    'all-of-router',
    'allocate',
    'allow',
    'allow-connections',
    'allow-default',
    'allow-fail-through',
    'allow-nopassword-remote-login',
    'allow-self-ping',
    'allow-v4mapped-packets',
    'allowas-in',
    'allowed',
    'alternate-address',
    'always',
    'always-compare-med',
    'always-on',
    'always-on-vpn',
    'always-send',
    'am-disable',
    'am-scan-profile',
    'amon',
    'and',
    'antenna',
    'any',
    'any-ipv4',
    'any-ipv6',
    'any-remote-host',
    'any-service',
    'any4',
    'any6',
    'anyconnect',
    'anyconnect-essentials',
    'aol',
    'ap',
    'ap-blacklist-time',
    'ap-classification-rule',
    'ap-crash-transfer',
    'ap-group',
    'ap-lacp-striping-ip',
    'ap-name',
    'ap-rule-matching',
    'ap-system-profile',
    'api',
    'app',
    'appcategory',
    'appletalk',
    'application',
    'application-protocol',
    'application-set',
    'application-tracking',
    'applications',
    'apply',
    'apply-macro',
    'apply-path',
    'aqm-register-fnf',
    'archive',
    'archive-length',
    'archive-size',
    'area',
    'area-password',
    'area-range',
    'arm-profile',
    'arm-rf-domain-profile',
    'arp',
    'arp-resp',
    'as',
    'as-number',
    'as-override',
    'as-path',
    'as-path-expand',
    'as-path-prepend',
    'as-path-set',
    'as-set',
    'asa',
    'ascii-authentication',
    'ascii-text',
    'asdm',
    'asdm-buffer-size',
    'asdot-notation',
    'asn',
    'assembler',
    'assignment',
    'assoc-retransmit',
    'associate',
    'association',
    'asymmetric',
    'async',
    'async-bootp',
    'asynchronous',
    'atm',
    'attached',
    'attribute',
    'attribute-download',
    'attribute-map',
    'attribute-names',
    'attribute-set',
    'attributes',
    'audit',
    'auth',
    'auth-failure-blacklist-time',
    'auth-port',
    'auth-proxy',
    'auth-server',
    'authenticate',
    'authentication',
    'authentication-algorithm',
    'authentication-dot1x',
    'authentication-key',
    'authentication-key-chain',
    'authentication-key-chains',
    'authentication-mac',
    'authentication-method',
    'authentication-order',
    'authentication-port',
    'authentication-retries',
    'authentication-server',
    'authentication-server-group',
    'authentication-type',
    'authoritative',
    'authorization',
    'authorization-required',
    'authorization-server-group',
    'authtype',
    'auto',
    'auto-cert-allow-all',
    'auto-cert-allowed-addrs',
    'auto-cert-prov',
    'auto-cost',
    'auto-export',
    'auto-import',
    'auto-local-addr',
    'auto-negotiation',
    'auto-recovery',
    'auto-rp',
    'auto-shutdown-new-neighbors',
    'auto-snapshot',
    'auto-summary',
    'auto-sync',
    'auto-tunnel',
    'auto-update',
    'auto-upgrade',
    'autohang',
    'autohangup',
    'autonomous-system',
    'autorecovery',
    'autoroute',
    'autorp',
    'autoselect',
    'autostate',
    'aux',
    'back-up',
    'backbonefast',
    'background-routes-enable',
    'backoff-time',
    'backup',
    'backup-router',
    'backupcrf',
    'band-steering',
    'bandwidth',
    'bandwidth-contract',
    'bandwidth-percentage',
    'banner',
    'bash',
    'basic',
    'batfish_role_declarations',
    'bcmc-optimization',
    'bcn-rpt-req-profile',
    'beacon',
    'bestpath',
    'beyond-scope',
    'bfd',
    'bfd-enable',
    'bfd-liveness-detection',
    'bgp',
    'bgp-community',
    'bgp-policy',
    'bidir-enable',
    'bidir-offer-interval',
    'bidir-offer-limit',
    'bidir-rp-limit',
    'biff',
    'bind',
    'bind-interface',
    'bittorrent',
    'bittorrent-application',
    'bkup-lms-ip',
    'blackhole',
    'blacklist',
    'blacklist-time',
    'block',
    'bloggerd',
    'bmp',
    'bonddevs',
    'bonding',
    'bondmiimon',
    'bondmode',
    'bool',
    'boot',
    'boot-end-marker',
    'boot-server',
    'boot-start-marker',
    'bootfile',
    'bootp',
    'bootp-relay',
    'bootpc',
    'bootps',
    'border',
    'border-router',
    'both',
    'boundary',
    'bpdufilter',
    'bpduguard',
    'bridge',
    'bridge-domain',
    'bridge-domains',
    'bridge-group',
    'bridge-priority',
    'broadcast',
    'broadcast-address',
    'broadcast-client',
    'broadcast-filter',
    'bsd-client',
    'bsd-username',
    'bsr-border',
    'bsr-candidate',
    'buckets',
    'buffer-size',
    'buffered',
    'buffers',
    'building configuration',
    'bundle',
    'burst-size',
    'ca',
    'cable-downstream',
    'cable-range',
    'cable-upstream',
    'cablelength',
    'cache',
    'cache-timeout',
    'call',
    'call-block',
    'call-forward',
    'call-home',
    'call-manager-fallback',
    'caller-id',
    'callhome',
    'cam-acl',
    'cam-profile',
    'capability',
    'captive',
    'captive-portal-cert',
    'capture',
    'card',
    'card-trap-inh',
    'carrier-delay',
    'cas-custom',
    'case',
    'categories',
    'ccc',
    'ccm',
    'ccm-group',
    'ccm-manager',
    'cdp',
    'cdp-url',
    'cef',
    'centralized-licensing-enable',
    'certificate',
    'certificates',
    'cfs',
    'cgmp',
    'channel',
    'channel-group',
    'channel-protocol',
    'channelized',
    'chap',
    'chargen',
    'chassis',
    'chassis-id',
    'chat-script',
    'check',
    'cipc',
    'cir',
    'circuit-type',
    'cisp',
    'citrix-ica',
    'class',
    'class-map',
    'class-of-service',
    'classless',
    'cleanup',
    'clear',
    'cli',
    'client',
    'client-group',
    'client-identifier',
    'client-list-name',
    'client-name',
    'client-to-client',
    'clients',
    'clns',
    'clock',
    'clock-period',
    'closed',
    'cluster',
    'cluster-id',
    'cmd',
    'cmts',
    'cns',
    'coap',
    'codec',
    'codes',
    'collect',
    'collect-stats',
    'color',
    'color2',
    'comm-list',
    'commander-address',
    'commands',
    'commit',
    'common',
    'community',
    'community-list',
    'community-map',
    'community-set',
    'compare-routerid',
    'compatible',
    'compress-configuration-files',
    'compression',
    'con',
    'condition',
    'conf-level-incr',
    'confdconfig',
    'confederation',
    'config',
    'config-commands',
    'config-management',
    'config-register',
    'configuration',
    'configure',
    'configversion',
    'conflict-policy',
    'conform-action',
    'congestion-control',
    'conn',
    'conn-holddown',
    'connect-retry',
    'connect-source',
    'connected',
    'connection',
    'connection-reuse',
    'connection-type',
    'connections',
    'connections-limit',
    'console',
    'contact',
    'contact-email-addr',
    'contact-name',
    'context',
    'contract-id',
    'control-apps-use-mgmt-port',
    'control-plane',
    'control-plane-security',
    'control-word',
    'controller',
    'controller-ip',
    'conversion-error',
    'cookie',
    'copp',
    'cops',
    'copy',
    'cos',
    'cos-mapping',
    'cos-next-hop-map',
    'cos-queue-group',
    'cost',
    'count',
    'counter',
    'counters',
    'country',
    'country-code',
    'cpd',
    'cptone',
    'cpu-share',
    'crashinfo',
    'crc',
    'credentials',
    'credibility-protocol-preference',
    'critical',
    'crl',
    'crypto',
    'crypto-local',
    'cryptochecksum',
    'cryptographic-algorithm',
    'cs1',
    'cs2',
    'cs3',
    'cs4',
    'cs5',
    'cs6',
    'cs7',
    'csd',
    'csnp-interval',
    'ctiqbe',
    'ctl-file',
    'cts',
    'current configuration',
    'custom',
    'customer-id',
    'cvspserver',
    'cvx',
    'cwr',
    'd20-ggrp-default',
    'd30-ggrp-default',
    'daemon',
    'dampen',
    'dampen-igp-metric',
    'dampening',
    'damping',
    'data-privacy',
    'database',
    'databits',
    'daytime',
    'dbl',
    'dcb',
    'dcb-buffer-threshold',
    'dcb-policy',
    'dcbx',
    'dce-mode',
    'ddos-protection',
    'deactivate',
    'dead-interval',
    'dead-peer-detection',
    'deadtime',
    'debug',
    'debug-trace',
    'debugging',
    'decap-group',
    'default-action',
    'default-address-selection',
    'default-cost',
    'default-destination',
    'default-domain',
    'default-gateway',
    'default-group-policy',
    'default-guest-role',
    'default-gw',
    'default-information',
    'default-information-originate',
    'default-inspection-traffic',
    'default-lsa',
    'default-max-frame-size',
    'default-metric',
    'default-network',
    'default-originate',
    'default-policy',
    'default-role',
    'default-router',
    'default-taskgroup',
    'default-tos-qos10',
    'default-value',
    'defaults',
    'definition',
    'del',
    'delay',
    'delay-start',
    'delete',
    'delete-dynamic-learn',
    'delimiter',
    'demand-circuit',
    'dense-mode',
    'deny',
    'deny-all',
    'deny-inter-user-traffic',
    'depi',
    'depi-class',
    'depi-tunnel',
    'deploy',
    'derivation-rules',
    'des',
    'des-cbc',
    'des-sha1',
    'description',
    'desirable',
    'despassword',
    'dest-ip',
    'destination',
    'destination-address',
    'destination-host-unknown',
    'destination-ip',
    'destination-network-unknown',
    'destination-pattern',
    'destination-port',
    'destination-port-except',
    'destination-prefix-list',
    'destination-profile',
    'destination-slot',
    'destination-unreachable',
    'destination-vrf',
    'detail',
    'detect-adhoc-network',
    'detect-ap-flood',
    'detect-ap-impersonation',
    'detect-bad-wep',
    'detect-beacon-wrong-channel',
    'detect-chopchop-attack',
    'detect-client-flood',
    'detect-cts-rate-anomaly',
    'detect-eap-rate-anomaly',
    'detect-hotspotter',
    'detect-ht-40mhz-intolerance',
    'detect-ht-greenfield',
    'detect-invalid-address-combination',
    'detect-invalid-mac-oui',
    'detect-malformed-association-request',
    'detect-malformed-auth-frame',
    'detect-malformed-htie',
    'detect-malformed-large-duration',
    'detect-misconfigured-ap',
    'detect-overflow-eapol-key',
    'detect-overflow-ie',
    'detect-rate-anomalies',
    'detect-rts-rate-anomaly',
    'detect-tkip-replay-attack',
    'detect-valid-ssid-misuse',
    'detect-wireless-bridge',
    'detect-wireless-hosted-network',
    'deterministic-med',
    'dev',
    'device',
    'device-sensor',
    'df',
    'df-bit',
    'dfa-regex',
    'dfs',
    'dh-group',
    'dh-group14',
    'dh-group15',
    'dh-group16',
    'dh-group17',
    'dh-group18',
    'dh-group19',
    'dh-group2',
    'dh-group20',
    'dh-group21',
    'dh-group22',
    'dh-group23',
    'dh-group24',
    'dh-group25',
    'dh-group26',
    'dh-group5',
    'dhcp',
    'dhcp-giaddr',
    'dhcp-relay',
    'dhcpd',
    'dhcprelay',
    'diagnostic',
    'diagnostic-signature',
    'dial-control-mib',
    'dial-peer',
    'dial-string',
    'dialer',
    'dialer-group',
    'dialer-list',
    'dialplan-pattern',
    'dialplan-profile',
    'direct',
    'direct-inward-dial',
    'directed-broadcast',
    'directed-request',
    'disable',
    'disable-4byte-as',
    'disable-advertisement',
    'disable-connected-check',
    'disable-peer-as-check',
    'discard',
    'discard-route',
    'discovered-ap-cnt',
    'discovery',
    'discriminator',
    'dispute',
    'distance',
    'distribute',
    'distribute-list',
    'distribution',
    'dm-fallback',
    'dns',
    'dns-domain',
    'dns-guard',
    'dns-server',
    'dns1',
    'dns2',
    'dnsix',
    'do',
    'docsis-enable',
    'docsis-group',
    'docsis-policy',
    'docsis-version',
    'docsis30-enable',
    'domain',
    'domain-id',
    'domain-list',
    'domain-lookup',
    'domain-name',
    'domain-search',
    'done',
    'dont-capability-negotiate',
    'dos-profile',
    'dot11',
    'dot11a-radio-profile',
    'dot11g-radio-profile',
    'dot11k-profile',
    'dot11r-profile',
    'dot1p-priority',
    'dot1q-tunnel',
    'dot1x-default-role',
    'dot1x-enable',
    'dot1x-server-group',
    'downlink',
    'downstream',
    'downstream-start-threshold',
    'dpg',
    'dr-priority',
    'drop',
    'ds-hello-interval',
    'ds-max-burst',
    'ds0-group',
    'dsa-signatures',
    'dscp',
    'dscp-value',
    'dsg',
    'dsl',
    'dsp',
    'dspfarm',
    'dsrwait',
    'dss',
    'dst-nat',
    'dstopts',
    'dsu',
    'dtmf-relay',
    'dual-active',
    'dual-mode-default-vlan',
    'dummy',
    'dump-on-panic',
    'duplex',
    'duplicate-message',
    'duration',
    'dvmrp',
    'dynamic',
    'dynamic-access-policy-record',
    'dynamic-capability',
    'dynamic-extended',
    'dynamic-map',
    'dynamic-mcast-optimization',
    'dynamic-mcast-optimization-thresh',
    'e164',
    'e164-pattern-map',
    'eap-passthrough',
    'eapol-rate-opt',
    'early-offer',
    'ebgp',
    'ebgp-multihop',
    'ece',
    'echo',
    'echo-cancel',
    'echo-reply',
    'echo-request',
    'ecmp',
    'ecmp-group',
    'ecn',
    'edca-parameters-profile',
    'edge',
    'ef',
    'egp',
    'egress',
    'egress-interface-selection',
    'eibgp',
    'eigrp',
    'eklogin',
    'ekshell',
    'else',
    'elseif',
    'email',
    'email-addr',
    'email-contact',
    'emergencies',
    'empty',
    'emptypath',
    'enable',
    'enable-acl-cam-sharing',
    'enable-acl-counter',
    'enable-authentication',
    'enable-qos-statistics',
    'enable-welcome-page',
    'enabled',
    'encapsulation',
    'encr',
    'encrypted',
    'encrypted-password',
    'encryption',
    'encryption-algorithm',
    'end',
    'end-class-map',
    'end-policy',
    'end-policy-map',
    'end-set',
    'endif',
    'enet-link-profile',
    'enforce-dhcp',
    'enforce-first-as',
    'enforce-rule',
    'enforced',
    'engine',
    'enhanced-hash-key',
    'enrollment',
    'environment',
    'environment-monitor',
    'eof',
    'eq',
    'errdisable',
    'error',
    'error-enable',
    'error-passthru',
    'error-rate-threshold',
    'error-recovery',
    'errors',
    'erspan-id',
    'escape-character',
    'esm',
    'esp',
    'esp-3des',
    'esp-aes',
    'esp-des',
    'esp-gcm',
    'esp-gmac',
    'esp-group',
    'esp-md5-hmac',
    'esp-null',
    'esp-seal',
    'esp-sha-hmac',
    'esp-sha256-hmac',
    'essid',
    'establish-tunnels',
    'established',
    'eth',
    'ether-options',
    'etherchannel',
    'ethernet',
    'ethernet-services',
    'ethernet-switching',
    'ethernet-switching-options',
    'etype',
    'evaluate',
    'event',
    'event-handler',
    'event-history',
    'event-options',
    'event-thresholds-profile',
    'events',
    'exact',
    'exceed-action',
    'except',
    'exception',
    'exception-slave',
    'exclude',
    'excluded-address',
    'exec',
    'exec-timeout',
    'execute',
    'exit',
    'exit-address-family',
    'exit-peer-policy',
    'exit-peer-session',
    'exit-vrf',
    'exp',
    'expanded',
    'expect',
    'explicit-null',
    'explicit-priority',
    'export',
    'export-protocol',
    'export-rib',
    'exporter',
    'exporter-map',
    'expression',
    'extcomm-list',
    'extcommunity',
    'extcommunity-list',
    'extend',
    'extended',
    'extended-counters',
    'extended-delay',
    'extensions',
    'external',
    'external-interface',
    'external-lsa',
    'external-preference',
    'external-router-id',
    'fabric',
    'fabric-mode',
    'fabric-options',
    'fabricpath',
    'facility',
    'facility-alarm',
    'facility-override',
    'fail-filter',
    'fail-over',
    'failed',
    'failed-list',
    'failover',
    'fair-queue',
    'fall-over',
    'fallback',
    'fallback-dn',
    'false',
    'family',
    'fan',
    'fast-age',
    'fast-detect',
    'fast-external-fallover',
    'fast-flood',
    'fast-reroute',
    'fastether-options',
    'fax',
    'fax-relay',
    'fcoe',
    'fdl',
    'feature',
    'feature-set',
    'fex',
    'fex-fabric',
    'fiber-node',
    'fields',
    'file',
    'file-browsing',
    'file-entry',
    'file-size',
    'file-transfer',
    'filter',
    'filter-list',
    'fin',
    'finger',
    'firewall',
    'firewall-visibility',
    'firmware',
    'first-fragment',
    'flap-list',
    'flash',
    'flash-override',
    'flexible-vlan-tagging',
    'floating-conn',
    'flow',
    'flow-accounting',
    'flow-aggregation',
    'flow-cache',
    'flow-capture',
    'flow-control',
    'flow-export',
    'flow-sampler',
    'flow-sampler-map',
    'flow-sampling-mode',
    'flow-top-talkers',
    'flowcont',
    'flowcontrol',
    'flush-at-activation',
    'flush-r1-on-new-r0',
    'for',
    'force',
    'force-up',
    'forever',
    'format',
    'forward',
    'forward-digits',
    'forward-protocol',
    'forward-snooped-clients',
    'forwarder',
    'forwarding',
    'forwarding-class',
    'forwarding-options',
    'forwarding-table',
    'fpd',
    'fqdn',
    'fragment',
    'fragment-offset',
    'fragment-offset-except',
    'fragmentation-needed',
    'fragments',
    'frame-relay',
    'framing',
    'free-channel-index',
    'frequency',
    'fri',
    'from',
    'from-zone',
    'ft',
    'ftp',
    'ftp-data',
    'ftp-server',
    'full-duplex',
    'full-txt',
    'g709',
    'g729',
    'gatekeeper',
    'gateway',
    'gateway1',
    'gbps',
    'ge',
    'general-group-defaults',
    'general-profile',
    'generate',
    'gid',
    'gig-default',
    'gigether-options',
    'glbp',
    'global',
    'global-mtu',
    'global-port-security',
    'globalenforcepriv',
    'gopher',
    'graceful-restart',
    'gracetime',
    'grant',
    'gratuitous-arps',
    'gre',
    'green',
    'group',
    'group-alias',
    'group-ike-id',
    'group-list',
    'group-lock',
    'group-object',
    'group-policy',
    'group-range',
    'group-timeout',
    'group-url',
    'group1',
    'group14',
    'group2',
    'group5',
    'groups',
    'gt',
    'guaranteed',
    'guard',
    'guest-access-email',
    'guest-logon',
    'guest-mode',
    'gui',
    'guimenuname',
    'gw-type-prefix',
    'h225',
    'h323',
    'h323-gateway',
    'ha-policy',
    'half-closed',
    'half-duplex',
    'handover-trigger-profile',
    'hardware',
    'hardware-address',
    'hardware-count',
    'hash',
    'hash-algorithm',
    'hash-key',
    'header-compression',
    'header-passing',
    'heartbeat-interval',
    'heartbeat-time',
    'hello',
    'hello-authentication-key',
    'hello-authentication-type',
    'hello-interval',
    'hello-multiplier',
    'hello-padding',
    'hello-password',
    'helper-address',
    'helpers',
    'hidden',
    'hidden-shares',
    'hidekeys',
    'high',
    'high-availability',
    'high-resolution',
    'history',
    'hmac-md5-96',
    'hmac-sha-1',
    'hmac-sha-1-96',
    'hmac-sha1-96',
    'hold-queue',
    'hold-time',
    'homedir',
    'hop-by-hop',
    'hop-limit',
    'hoplimit',
    'hops-of-statistics-kept',
    'host',
    'host-association',
    'host-inbound-traffic',
    'host-info',
    'host-name',
    'host-proxy',
    'host-routing',
    'host-unknown',
    'host-unreachable',
    'hostname',
    'hotspot',
    'hpm',
    'hsrp',
    'ht-ssid-profile',
    'http',
    'http-commands',
    'https',
    'hunt',
    'hw-id',
    'hw-module',
    'hw-switch',
    'ibgp',
    'iccp',
    'icmp',
    'icmp-code',
    'icmp-echo',
    'icmp-error',
    'icmp-errors',
    'icmp-object',
    'icmp-type',
    'icmp6',
    'icmp6-code',
    'icmp6-type',
    'icmpv6',
    'id',
    'id-mismatch',
    'id-randomization',
    'ideal-coverage-index',
    'ident',
    'ident-reset',
    'identity',
    'idle',
    'idle-timeout',
    'idletimeout',
    'idp-cert',
    'ids',
    'ids-profile',
    'iec',
    'ietf-format',
    'if',
    'if-needed',
    'ifacl',
    'ifdescr',
    'ifindex',
    'ifmap',
    'ifmib',
    'ifname',
    'igmp',
    'igmp-snooping',
    'ignore',
    'ignore-l3-incompletes',
    'igp',
    'igp-cost',
    'igrp',
    'ike',
    'ike-esp-nat',
    'ike-group',
    'ike-policy',
    'ike-user-type',
    'ikev1',
    'ikev2',
    'ikev2-profile',
    'ikev2-reauth',
    'ilmi-keepalive',
    'imap',
    'imap4',
    'immediate',
    'immediately',
    'impersonation-profile',
    'implicit-user',
    'import',
    'import-policy',
    'import-rib',
    'inactive',
    'inactive:',
    'inactivity-timeout',
    'inactivity-timer',
    'inband',
    'inbound',
    'include-mp-next-hop',
    'include-stub',
    'incoming',
    'incomplete',
    'index',
    'indirect-next-hop',
    'indirect-next-hop-change-acknowledgements',
    'inet',
    'inet-mdt',
    'inet-mvpn',
    'inet-vpn',
    'inet6',
    'inet6-vpn',
    'infinity',
    'information',
    'information-reply',
    'information-request',
    'informational',
    'ingress',
    'inherit',
    'inheritance-disable',
    'init',
    'init-string',
    'init-tech-list',
    'initial-role',
    'initiate',
    'inner',
    'input',
    'input-list',
    'input-vlan-map',
    'inservice',
    'inside',
    'inspect',
    'install',
    'install-nexthop',
    'instance',
    'instance-type',
    'integer',
    'integrity',
    'inter',
    'interactive-commands',
    'interarea',
    'intercept',
    'interconnect-device',
    'interface',
    'interface-mode',
    'interface-routes',
    'interface-specific',
    'interface-switch',
    'interface-transmit-statistics',
    'interface-type',
    'interfaces',
    'internal',
    'internet',
    'internet-options',
    'interval',
    'interworking',
    'intra',
    'invalid-spi-recovery',
    'invalid-username-log',
    'invert',
    'invert-match',
    'ios-regex',
    'ip',
    'ip route table for vrf',
    'ip-address',
    'ip-flow-export-profile',
    'ip-options',
    'ipaddr',
    'ipaddress',
    'ipbroadcast',
    'ipc',
    'ipenacl',
    'iphc-format',
    'ipinip',
    'ipip',
    'ipmask',
    'ipsec',
    'ipsec-interfaces',
    'ipsec-isakmp',
    'ipsec-over-tcp',
    'ipsec-policy',
    'ipsec-proposal',
    'ipsec-udp',
    'ipsec-vpn',
    'ipsla',
    'iptables',
    'ipv4',
    'ipv4-l5',
    'ipv6',
    'ipv6-address-pool',
    'ipx',
    'irc',
    'irdp',
    'is',
    'is directly connected',
    'is variably subnetted',
    'is-fragment',
    'is-local',
    'is-type',
    'isakmp',
    'isdn',
    'isis',
    'isis-enhanced',
    'isis-metric',
    'isl',
    'iso',
    'isolate',
    'ispf',
    'issuer-name',
    'iuc',
    'join-group',
    'jumbo',
    'jumbomtu',
    'junos-aol',
    'junos-bgp',
    'junos-biff',
    'junos-bootpc',
    'junos-bootps',
    'junos-chargen',
    'junos-cifs',
    'junos-cvspserver',
    'junos-dhcp-client',
    'junos-dhcp-relay',
    'junos-dhcp-server',
    'junos-discard',
    'junos-dns-tcp',
    'junos-dns-udp',
    'junos-echo',
    'junos-finger',
    'junos-ftp',
    'junos-gnutella',
    'junos-gopher',
    'junos-gre',
    'junos-gtp',
    'junos-h323',
    'junos-host',
    'junos-http',
    'junos-http-ext',
    'junos-https',
    'junos-icmp-all',
    'junos-icmp-ping',
    'junos-icmp6-all',
    'junos-icmp6-dst-unreach-addr',
    'junos-icmp6-dst-unreach-admin',
    'junos-icmp6-dst-unreach-beyond',
    'junos-icmp6-dst-unreach-port',
    'junos-icmp6-dst-unreach-route',
    'junos-icmp6-echo-reply',
    'junos-icmp6-echo-request',
    'junos-icmp6-packet-to-big',
    'junos-icmp6-param-prob-header',
    'junos-icmp6-param-prob-nexthdr',
    'junos-icmp6-param-prob-option',
    'junos-icmp6-time-exceed-reassembly',
    'junos-icmp6-time-exceed-transit',
    'junos-ident',
    'junos-ike',
    'junos-ike-nat',
    'junos-imap',
    'junos-imaps',
    'junos-internet-locator-service',
    'junos-irc',
    'junos-l2tp',
    'junos-ldap',
    'junos-ldp-tcp',
    'junos-ldp-udp',
    'junos-lpr',
    'junos-mail',
    'junos-mgcp',
    'junos-mgcp-ca',
    'junos-mgcp-ua',
    'junos-ms-rpc',
    'junos-ms-rpc-any',
    'junos-ms-rpc-epm',
    'junos-ms-rpc-iis-com',
    'junos-ms-rpc-iis-com-1',
    'junos-ms-rpc-iis-com-adminbase',
    'junos-ms-rpc-msexchange',
    'junos-ms-rpc-msexchange-directory-nsp',
    'junos-ms-rpc-msexchange-directory-rfr',
    'junos-ms-rpc-msexchange-info-store',
    'junos-ms-rpc-tcp',
    'junos-ms-rpc-udp',
    'junos-ms-rpc-uuid-any-tcp',
    'junos-ms-rpc-uuid-any-udp',
    'junos-ms-rpc-wmic',
    'junos-ms-rpc-wmic-admin',
    'junos-ms-rpc-wmic-admin2',
    'junos-ms-rpc-wmic-mgmt',
    'junos-ms-rpc-wmic-webm-callresult',
    'junos-ms-rpc-wmic-webm-classobject',
    'junos-ms-rpc-wmic-webm-level1login',
    'junos-ms-rpc-wmic-webm-login-clientid',
    'junos-ms-rpc-wmic-webm-login-helper',
    'junos-ms-rpc-wmic-webm-objectsink',
    'junos-ms-rpc-wmic-webm-refreshing-services',
    'junos-ms-rpc-wmic-webm-remote-refresher',
    'junos-ms-rpc-wmic-webm-services',
    'junos-ms-rpc-wmic-webm-shutdown',
    'junos-ms-sql',
    'junos-msn',
    'junos-nbds',
    'junos-nbname',
    'junos-netbios-session',
    'junos-nfs',
    'junos-nfsd-tcp',
    'junos-nfsd-udp',
    'junos-nntp',
    'junos-ns-global',
    'junos-ns-global-pro',
    'junos-nsm',
    'junos-ntalk',
    'junos-ntp',
    'junos-ospf',
    'junos-pc-anywhere',
    'junos-persistent-nat',
    'junos-ping',
    'junos-pingv6',
    'junos-pop3',
    'junos-pptp',
    'junos-printer',
    'junos-r2cp',
    'junos-radacct',
    'junos-radius',
    'junos-realaudio',
    'junos-rip',
    'junos-routing-inbound',
    'junos-rsh',
    'junos-rtsp',
    'junos-sccp',
    'junos-sctp-any',
    'junos-sip',
    'junos-smb',
    'junos-smb-session',
    'junos-smtp',
    'junos-snmp-agentx',
    'junos-snpp',
    'junos-sql-monitor',
    'junos-sqlnet-v1',
    'junos-sqlnet-v2',
    'junos-ssh',
    'junos-stun',
    'junos-sun-rpc',
    'junos-sun-rpc-any',
    'junos-sun-rpc-any-tcp',
    'junos-sun-rpc-any-udp',
    'junos-sun-rpc-mountd',
    'junos-sun-rpc-mountd-tcp',
    'junos-sun-rpc-mountd-udp',
    'junos-sun-rpc-nfs',
    'junos-sun-rpc-nfs-access',
    'junos-sun-rpc-nfs-tcp',
    'junos-sun-rpc-nfs-udp',
    'junos-sun-rpc-nlockmgr',
    'junos-sun-rpc-nlockmgr-tcp',
    'junos-sun-rpc-nlockmgr-udp',
    'junos-sun-rpc-portmap',
    'junos-sun-rpc-portmap-tcp',
    'junos-sun-rpc-portmap-udp',
    'junos-sun-rpc-rquotad',
    'junos-sun-rpc-rquotad-tcp',
    'junos-sun-rpc-rquotad-udp',
    'junos-sun-rpc-ruserd',
    'junos-sun-rpc-ruserd-tcp',
    'junos-sun-rpc-ruserd-udp',
    'junos-sun-rpc-sadmind',
    'junos-sun-rpc-sadmind-tcp',
    'junos-sun-rpc-sadmind-udp',
    'junos-sun-rpc-sprayd',
    'junos-sun-rpc-sprayd-tcp',
    'junos-sun-rpc-sprayd-udp',
    'junos-sun-rpc-status',
    'junos-sun-rpc-status-tcp',
    'junos-sun-rpc-status-udp',
    'junos-sun-rpc-tcp',
    'junos-sun-rpc-udp',
    'junos-sun-rpc-walld',
    'junos-sun-rpc-walld-tcp',
    'junos-sun-rpc-walld-udp',
    'junos-sun-rpc-ypbind',
    'junos-sun-rpc-ypbind-tcp',
    'junos-sun-rpc-ypbind-udp',
    'junos-sun-rpc-ypserv',
    'junos-sun-rpc-ypserv-tcp',
    'junos-sun-rpc-ypserv-udp',
    'junos-syslog',
    'junos-tacacs',
    'junos-tacacs-ds',
    'junos-talk',
    'junos-tcp-any',
    'junos-telnet',
    'junos-tftp',
    'junos-udp-any',
    'junos-uucp',
    'junos-vdo-live',
    'junos-vnc',
    'junos-wais',
    'junos-who',
    'junos-whois',
    'junos-winframe',
    'junos-wxcontrol',
    'junos-x-windows',
    'junos-xnm-clear-text',
    'junos-xnm-ssl',
    'junos-ymsg',
    'kbps',
    'keep',
    'keepalive',
    'keepalive-enable',
    'keepout',
    'kerberos',
    'kerberos-sec',
    'kernel',
    'key',
    'key-chain',
    'key-exchange',
    'key-hash',
    'key-source',
    'key-string',
    'keypair',
    'keypath',
    'keyring',
    'keys',
    'keystore',
    'klogin',
    'kpasswd',
    'krb-prop',
    'krbupdate',
    'kron',
    'kshell',
    'l',
    'l2',
    'l2-filter',
    'l2-interface',
    'l2-src',
    'l2circuit',
    'l2protocol',
    'l2protocol-tunnel',
    'l2tp',
    'l2tp-class',
    'l2tpv3',
    'l2vpn',
    'l3-interface',
    'label',
    'label-switched-path',
    'labeled-unicast',
    'lacp',
    'lacp-timeout',
    'lag',
    'lan',
    'lane',
    'lapb',
    'last-as',
    'last-member-query-count',
    'last-member-query-interval',
    'last-member-query-response-time',
    'layer2-control',
    'lcd-menu',
    'ldap',
    'ldap-base-dn',
    'ldap-login',
    'ldap-login-dn',
    'ldap-naming-attribute',
    'ldap-scope',
    'ldaps',
    'ldp',
    'ldp-synchronization',
    'le',
    'learn-vlan-1p-priority',
    'lease',
    'length',
    'level',
    'level-1',
    'level-1-2',
    'level-2',
    'level-2-only',
    'license',
    'life',
    'lifetime',
    'lifetime-kilobytes',
    'lifetime-seconds',
    'limit',
    'limit-dn',
    'limit-resource',
    'line',
    'linecard',
    'linecard-group',
    'linecode',
    'link',
    'link-fail',
    'link-fault-signaling',
    'link-mode',
    'link-protection',
    'link-type',
    'linkdebounce',
    'list',
    'listen',
    'listen-port',
    'listener',
    'lldp',
    'lldp-med',
    'lms-ip',
    'lms-preemption',
    'lo',
    'load-balance',
    'load-balancing',
    'load-interval',
    'load-sharing',
    'local',
    'local-address',
    'local-as',
    'local-identity',
    'local-interface',
    'local-ip',
    'local-port',
    'local-preference',
    'local-v6-addr',
    'local-volatile',
    'locale',
    'localip',
    'location',
    'log',
    'log-adj-changes',
    'log-adjacency-changes',
    'log-enable',
    'log-input',
    'log-internal-sync',
    'log-neighbor-changes',
    'log-prefix',
    'log-updown',
    'logfile',
    'logging',
    'logical-systems',
    'login',
    'login-attempts',
    'login-authentication',
    'login-page',
    'loginsession',
    'logout-warning',
    'longer',
    'lookup',
    'loopback',
    'loopguard',
    'loops',
    'loss-priority',
    'lotusnotes',
    'low',
    'lpd',
    'lpts',
    'lre',
    'lrq',
    'lsp',
    'lsp-equal-cost',
    'lsp-gen-interval',
    'lsp-interval',
    'lsp-lifetime',
    'lsp-password',
    'lsp-refresh-interval',
    'lsping',
    'lt',
    'mac',
    'mac-address',
    'mac-address-table',
    'mac-default-role',
    'mac-learn',
    'mac-server-group',
    'machine-authentication',
    'macro',
    'mail-server',
    'main',
    'main-cpu',
    'maintenance',
    'management',
    'management-access',
    'management-only',
    'management-plane',
    'management-profile',
    'manager',
    'mangle',
    'map',
    'map-class',
    'map-group',
    'map-list',
    'mapped-port',
    'mapping',
    'martians',
    'mask',
    'mask-reply',
    'mask-request',
    'master',
    'master-only',
    'masterip',
    'match',
    'match-all',
    'match-any',
    'matches-any',
    'matches-every',
    'max-associations',
    'max-authentication-failures',
    'max-burst',
    'max-clients',
    'max-concat-burst',
    'max-conferences',
    'max-configuration-rollbacks',
    'max-configurations-on-flash',
    'max-connections',
    'max-dn',
    'max-ephones',
    'max-ifindex-per-module',
    'max-lsa',
    'max-lsp-lifetime',
    'max-metric',
    'max-rate',
    'max-route',
    'max-session-number',
    'max-sessions',
    'max-tx-power',
    'maxconnections',
    'maximum',
    'maximum-accepted-routes',
    'maximum-labels',
    'maximum-paths',
    'maximum-peers',
    'maximum-prefix',
    'maximum-routes',
    'maxpoll',
    'maxstartups',
    'maxsubs',
    'mbps',
    'mbssid',
    'mcast-boundary',
    'mcast-rate-opt',
    'md5',
    'mdix',
    'mdt',
    'med',
    'media',
    'media-termination',
    'media-type',
    'medium',
    'medium-high',
    'medium-low',
    'member',
    'members',
    'memory',
    'memory-size',
    'menu',
    'menuname',
    'mesh-cluster-profile',
    'mesh-group',
    'mesh-ht-ssid-profile',
    'mesh-radio-profile',
    'message-counter',
    'message-digest-key',
    'message-length',
    'message-level',
    'message-size',
    'metering',
    'method',
    'method-utilization',
    'metric',
    'metric-out',
    'metric-style',
    'metric-type',
    'metric2',
    'mfib',
    'mfib-mode',
    'mfwd',
    'mgcp',
    'mgcp-ca',
    'mgcp-pat',
    'mgcp-ua',
    'mgmt-auth',
    'mgmt-server',
    'mgmt-user',
    'mh',
    'mib',
    'microcode',
    'microsoft-ds',
    'midcall-signaling',
    'min-packet-size',
    'min-rate',
    'min-tx-power',
    'minimal',
    'minimum',
    'minimum-interval',
    'minimum-links',
    'minpoll',
    'mirror',
    'mlag',
    'mld',
    'mld-query',
    'mld-reduction',
    'mld-report',
    'mldv2',
    'mls',
    'mobile',
    'mobile-ip',
    'mobile-redirect',
    'mobileip-agent',
    'mobilip-mn',
    'mobility',
    'mode',
    'modem',
    'modulation-profile',
    'module-type',
    'mon',
    'monitor',
    'monitor-interface',
    'monitor-map',
    'monitor-session',
    'monitoring',
    'monitoring-basics',
    'mop',
    'motd',
    'mpls',
    'mpls-label',
    'mroute',
    'mroute-cache',
    'ms-rpc',
    'mschap',
    'mschapv2',
    'msdp',
    'msdp-peer',
    'msie-proxy',
    'msrpc',
    'mst',
    'mstp',
    'mta',
    'mtu',
    'mtu-discovery',
    'mtu-ignore',
    'multi-chassis',
    'multi-config',
    'multi-topology',
    'multicast',
    'multicast-boundary',
    'multicast-mac',
    'multicast-routing',
    'multicast-static-only',
    'multihop',
    'multilink',
    'multipath',
    'multiple-as',
    'multiplier',
    'multiservice-options',
    'mvpn',
    'mvr',
    'name',
    'name-lookup',
    'name-resolution',
    'name-server',
    'named-key',
    'nameif',
    'names',
    'nameserver',
    'namespace',
    'nat',
    'nat-control',
    'nat-transparency',
    'nat-traversal',
    'native',
    'native-vlan-id',
    'natpool',
    'nbar',
    'nd',
    'nd-na',
    'nd-ns',
    'nd-type',
    'negotiate',
    'negotiated',
    'negotiation',
    'neighbor',
    'neighbor-advertisement',
    'neighbor-discovery',
    'neighbor-down',
    'neighbor-filter',
    'neighbor-group',
    'neighbor-solicit',
    'neq',
    'nested',
    'net',
    'net-unreachable',
    'netbios-dgm',
    'netbios-ns',
    'netbios-ss',
    'netbios-ssn',
    'netconf',
    'netdestination',
    'netdestination6',
    'netexthdr',
    'netmask',
    'netservice',
    'network-clock',
    'network-clock-participate',
    'network-clock-select',
    'network-delay',
    'network-domain',
    'network-object',
    'network-summary-export',
    'network-unknown',
    'network-unreachable',
    'never',
    'new-model',
    'newinfo',
    'next',
    'next-header',
    'next-hop',
    'next-hop-self',
    'next-hop-third-party',
    'next-ip',
    'next-ip6',
    'next-server',
    'next-table',
    'nexthop',
    'nexthop-attribute',
    'nexthop-list',
    'nexthop-self',
    'nexthop1',
    'nexthop2',
    'nfsd',
    'nhop-only',
    'nhrp',
    'nls',
    'nmsp',
    'nntp',
    'no l4r_shim',
    'no-active-backbone',
    'no-adjacency-down-notification',
    'no-advertise',
    'no-anti-replay',
    'no-auto-negotiation',
    'no-client-reflect',
    'no-ecmp-fast-reroute',
    'no-export',
    'no-flow-control',
    'no-install',
    'no-ipv4-routing',
    'no-nat-traversal',
    'no-neighbor-down-notification',
    'no-neighbor-learn',
    'no-nexthop-change',
    'no-prepend',
    'no-readvertise',
    'no-redirects',
    'no-redistribution',
    'no-resolve',
    'no-retain',
    'no-rfc-1583',
    'no-summaries',
    'no-summary',
    'no-traps',
    'noauth',
    'node',
    'node-device',
    'node-group',
    'noe',
    'nohangup',
    'non-broadcast',
    'non-client-nrt',
    'non-deterministic-med',
    'non500-isakmp',
    'nonegotiate',
    'nonstop-routing',
    'nos',
    'not',
    'not-advertise',
    'notiffacility',
    'notification',
    'notification-timer',
    'notifications',
    'notifpriority',
    'notify',
    'notify-filter',
    'notifyaddressname',
    'notifyaddressservice',
    'notifyaddressstate',
    'notifyservicename',
    'notifyserviceprotocol',
    'notifyserviceraw',
    'nsf',
    'nsr',
    'nssa',
    'nssa-external',
    'ntalk',
    'ntp',
    'ntpaddress',
    'ntpaltaddress',
    'ntpsourceinterface',
    'null',
    'nv',
    'oam',
    'object',
    'object-group',
    'octet',
    'octetstring',
    'off',
    'offset',
    'on',
    'on-failure',
    'on-passive',
    'on-startup',
    'on-success',
    'one',
    'one-out-of',
    'open',
    'openflow',
    'openvpn',
    'operation',
    'opmode',
    'ops',
    'optical-monitor',
    'optimization-profile',
    'optimize',
    'optimized',
    'option',
    'options',
    'or',
    'origin',
    'origin-id',
    'originate',
    'originator-id',
    'orlonger',
    'ospf',
    'ospf-external-type-1',
    'ospf-external-type-2',
    'ospf-inter-area',
    'ospf-intra-area',
    'ospf-nssa-type-1',
    'ospf-nssa-type-2',
    'ospf3',
    'ospfv3',
    'other-access',
    'oui',
    'out-delay',
    'out-of-band',
    'outauthtype',
    'outbound-acl-check',
    'outer',
    'output',
    'output-list',
    'output-vlan-map',
    'outside',
    'overload',
    'overload-control',
    'override',
    'override-metric',
    'overrides',
    'owner',
    'p2p',
    'package',
    'packet',
    'packet-capture-defaults',
    'packet-length',
    'packet-length-except',
    'packet-too-big',
    'packetcable',
    'packetsize',
    'pager',
    'pan',
    'pan-options',
    'param',
    'parameter-problem',
    'parameters',
    'parent',
    'parity',
    'parser',
    'participate',
    'pass',
    'passive',
    'passive-interface',
    'passive-only',
    'passphrase',
    'passwd',
    'password',
    'password-policy',
    'password-storage',
    'pat-xlate',
    'path',
    'path-count',
    'path-echo',
    'path-jitter',
    'path-option',
    'path-retransmit',
    'path-selection',
    'pathcost',
    'paths',
    'paths-of-statistics-kept',
    'pathsize',
    'pause',
    'payload-protocol',
    'pcanywhere-data',
    'pcanywhere-status',
    'pcp',
    'pcp-value',
    'pd-route-injection',
    'peakdetect',
    'peer',
    'peer-address',
    'peer-as',
    'peer-config-check-bypass',
    'peer-gateway',
    'peer-group',
    'peer-id-validate',
    'peer-keepalive',
    'peer-link',
    'peer-policy',
    'peer-session',
    'peer-switch',
    'peer-to-peer',
    'peer-unit',
    'penalty-period',
    'per-packet',
    'per-unit-scheduler',
    'percent',
    'perfect-forward-secrecy',
    'periodic',
    'periodic-inventory',
    'periodic-refresh',
    'permanent',
    'permission',
    'permit',
    'permit-all',
    'permit-hostdown',
    'persistent',
    'persistent-nat',
    'pfc',
    'pfs',
    'pgm',
    'phone-contact',
    'phone-number',
    'phone-proxy',
    'phy',
    'physical-layer',
    'physical-port',
    'pickup',
    'pim',
    'pim-auto-rp',
    'pim-sparse',
    'ping',
    'pinning',
    'pki',
    'plat',
    'platform',
    'pm',
    'poap',
    'poe',
    'point-to-multipoint',
    'point-to-point',
    'police',
    'policer',
    'policies',
    'policy',
    'policy-list',
    'policy-map',
    'policy-map-input',
    'policy-map-output',
    'policy-options',
    'policy-statement',
    'poll-interval',
    'pool',
    'pop2',
    'pop3',
    'port',
    'port-channel',
    'port-channel-protocol',
    'port-mirror',
    'port-mirroring',
    'port-mode',
    'port-name',
    'port-object',
    'port-overloading',
    'port-overloading-factor',
    'port-priority',
    'port-profile',
    'port-randomization',
    'port-security',
    'port-type',
    'port-unreachable',
    'portfast',
    'portmode',
    'ports',
    'pos',
    'postrotuing',
    'power',
    'power-level',
    'power-mgr',
    'power-monitor',
    'poweroff',
    'ppm',
    'pptp',
    'prc-interval',
    'pre-equalization',
    'pre-shared-key',
    'pre-shared-keys',
    'pre-shared-secret',
    'precedence',
    'precision-timers',
    'predictor',
    'preempt',
    'prefer',
    'preference',
    'preferred',
    'preferred-path',
    'prefix',
    'prefix-export-limit',
    'prefix-length',
    'prefix-length-range',
    'prefix-limit',
    'prefix-list',
    'prefix-list-filter',
    'prefix-policy',
    'prefix-set',
    'prerotuing',
    'prf',
    'pri-group',
    'primary',
    'primary-port',
    'primary-priority',
    'printer',
    'priority',
    'priority-cost',
    'priority-flow-control',
    'priority-force',
    'priority-mapping',
    'priority-queue',
    'priv',
    'privacy',
    'private',
    'private-as',
    'private-key',
    'private-vlan',
    'privilege',
    'privilege-mode',
    'proactive',
    'probe',
    'process',
    'process-max-time',
    'processes',
    'profile',
    'progress_ind',
    'prompt',
    'propagate',
    'proposal',
    'proposal-set',
    'proposals',
    'proprietary',
    'protect',
    'protect-ssid',
    'protect-tunnel',
    'protect-valid-sta',
    'protection',
    'protocol',
    'protocol-http',
    'protocol-object',
    'protocol-violation',
    'protocols',
    'provider-tunnel',
    'provision',
    'provisioning-profile',
    'proxy-arp',
    'proxy-identity',
    'proxy-server',
    'pruning',
    'pseudo-ethernet',
    'pseudo-information',
    'pseudowire',
    'pseudowire-class',
    'psh',
    'ptp',
    'pubkey-chain',
    'public-key',
    'pvc',
    'q931',
    'qos',
    'qos-group',
    'qos-mapping',
    'qos-policy',
    'qos-policy-output',
    'qos-sc',
    'qualified-next-hop',
    'query-interval',
    'query-max-response-time',
    'query-only',
    'query-timeout',
    'queue',
    'queue-buffers',
    'queue-length',
    'queue-limit',
    'queue-monitor',
    'queue-set',
    'quit',
    'r2cp',
    'radacct',
    'radius',
    'radius-accounting',
    'radius-acct',
    'radius-common-pw',
    'radius-interim-accounting',
    'radius-options',
    'radius-server',
    'radprimacctsecret',
    'radprimsecret',
    'radsecacctsecret',
    'radsecsecret',
    'random',
    'random-detect',
    'random-detect-label',
    'range',
    'ras',
    'rate-limit',
    'rate-mode',
    'rate-thresholds-profile',
    'raw',
    'rbacl',
    'rc4-sha1',
    'rcmd',
    'rcp',
    'rcv-queue',
    'rd',
    'reachable-via',
    'react',
    'reaction',
    'read-only',
    'read-only-password',
    'read-write',
    'readvertise',
    'real',
    'real-time-config',
    'realaudio',
    'reauthentication',
    'receive',
    'receive-window',
    'reconnect-interval',
    'record',
    'record-entry',
    'red',
    'redirect',
    'redirect-fqdn',
    'redirect-list',
    'redirect-pause',
    'redirects',
    'redistribute',
    'redistribute-internal',
    'redistributed-prefixes',
    'redundancy',
    'redundancy-group',
    'redundant-ether-options',
    'redundant-parent',
    'reference-bandwidth',
    'reflect',
    'reflexive-list',
    'register-rate-limit',
    'register-source',
    'registered',
    'regulatory-domain-profile',
    'reject',
    'relay-agent-option',
    'reload',
    'reload-delay',
    'reload-type',
    'remark',
    'remote',
    'remote-access',
    'remote-as',
    'remote-id',
    'remote-ip',
    'remote-port',
    'remote-ports',
    'remote-server',
    'remote-span',
    'remoteaccesslist',
    'remove',
    'remove-private',
    'removed',
    'reoptimize',
    'replace-as',
    'reply-to',
    'req-trans-policy',
    'request',
    'request-data-size',
    'require-wpa',
    'resolution',
    'resolve',
    'resource',
    'resource-pool',
    'resources',
    'respond',
    'responder',
    'response',
    'restrict',
    'restricted',
    'retain',
    'retransmit',
    'retransmit-interval',
    'retransmit-timeout',
    'retries',
    'retry',
    'return',
    'reverse-access',
    'reverse-path',
    'reverse-route',
    'reverse-ssh',
    'reverse-telnet',
    'revertive',
    'revision',
    'revocation-check',
    'rewrite',
    'rf',
    'rf-channel',
    'rf-power',
    'rf-shutdown',
    'rf-switch',
    'rfc-3576-server',
    'rfc1583',
    'rfc1583compatibility',
    'rib',
    'rib-group',
    'rib-groups',
    'rib-has-route',
    'rib-metric-as-external',
    'rib-metric-as-internal',
    'ring',
    'rip',
    'ripng',
    'rkinit',
    'rlogin',
    'rmon',
    'robustness-variable',
    'rogue-ap-aware',
    'role',
    'root',
    'root-authentication',
    'rotary',
    'route',
    'route-cache',
    'route-distinguisher',
    'route-filter',
    'route-map',
    'route-only',
    'route-policy',
    'route-reflector-client',
    'route-source',
    'route-target',
    'route-type',
    'routed',
    'router',
    'router-advertisement',
    'router-alert',
    'router-discovery',
    'router-id',
    'router-interface',
    'router-lsa',
    'router-solicit',
    'router-solicitation',
    'routing',
    'routing-instance',
    'routing-instances',
    'routing-options',
    'rp',
    'rp-address',
    'rp-announce-filter',
    'rp-candidate',
    'rp-list',
    'rpc-program-number',
    'rpf-check',
    'rpf-vector',
    'rpm',
    'rrm-ie-profile',
    'rsa',
    'rsa-signatures',
    'rsakeypair',
    'rsh',
    'rst',
    'rstp',
    'rsvp',
    'rtcp-inactivity',
    'rtp',
    'rtp-port',
    'rtr',
    'rtr-adv',
    'rtsp',
    'rule',
    'rule-name',
    'rule-set',
    'run',
    'rx',
    'rx-cos-slot',
    'rxspeed',
    'sa-filter',
    'same-security-traffic',
    'sample',
    'sampler',
    'sampler-map',
    'samples-of-history-kept',
    'sampling',
    'sap',
    'sat',
    'satellite',
    'satellite-fabric-link',
    'saved-core-context',
    'saved-core-files',
    'scale-factor',
    'scan-time',
    'scanning',
    'sccp',
    'sched-type',
    'schedule',
    'scheduler',
    'scheme',
    'scope',
    'scp',
    'scramble',
    'screen',
    'script',
    'scripting',
    'scripts',
    'sctp',
    'sdm',
    'sdr',
    'sdrowner',
    'secondary',
    'secondary-dialtone',
    'secret',
    'secure-mac-address',
    'secureid-udp',
    'security',
    'security-association',
    'security-level',
    'security-zone',
    'securityv3',
    'select',
    'selective',
    'self',
    'send',
    'send-community',
    'send-community-ebgp',
    'send-extended-community-ebgp',
    'send-label',
    'send-lifetime',
    'send-rp-announce',
    'send-rp-discovery',
    'sender',
    'sensor',
    'sequence',
    'sequence-nums',
    'serial-number',
    'serve',
    'serve-only',
    'server',
    'server-arp',
    'server-group',
    'server-private',
    'server-type',
    'serverfarm',
    'service',
    'service-class',
    'service-filter',
    'service-list',
    'service-module',
    'service-object',
    'service-policy',
    'service-queue',
    'service-type',
    'services',
    'session',
    'session-authorization',
    'session-disconnect-warning',
    'session-group',
    'session-id',
    'session-limit',
    'session-open-mode',
    'session-protection',
    'session-timeout',
    'set',
    'set-color',
    'set-overload-bit',
    'setup',
    'severity',
    'sflow',
    'sftp',
    'sg-expiry-timer',
    'sha',
    'sha-256',
    'sha-384',
    'sha1',
    'sha2-256-128',
    'sha256',
    'sha384',
    'sha512',
    'shapassword',
    'shape',
    'shared-ike-id',
    'shared-secondary-secret',
    'shared-secret',
    'shelfname',
    'shell',
    'short',
    'short-txt',
    'shortcuts',
    'shortstring',
    'show',
    'shutdown',
    'signal',
    'signaling',
    'signalled-bandwidth',
    'signalled-name',
    'signalling',
    'signature',
    'signature-matching-profile',
    'signature-profile',
    'signing',
    'simple',
    'single-connection',
    'single-router-mode',
    'single-topology',
    'sip',
    'sip-disconnect',
    'sip-invite',
    'sip-midcall-req-timeout',
    'sip-profiles',
    'sip-provisional-media',
    'sip-server',
    'sip-ua',
    'sip_media',
    'sips',
    'site-id',
    'site-to-site',
    'sla',
    'slot',
    'slot-table-cos',
    'small-hello',
    'smart-relay',
    'smp_affinity',
    'smtp',
    'smtp-server',
    'snmp',
    'snmp-authfail',
    'snmp-server',
    'snmp-trap',
    'snmpgetclient',
    'snmpgetcommunity',
    'snmpsourceinterface',
    'snmptrap',
    'snmptrapclient',
    'snmptrapcommunity',
    'snooping',
    'snp',
    'snpp',
    'snr-max',
    'snr-min',
    'sntp',
    'socks',
    'soft-preemption',
    'soft-reconfiguration',
    'software',
    'sonet',
    'sonet-options',
    'sort-by',
    'source',
    'source-address',
    'source-address-filter',
    'source-identity',
    'source-interface',
    'source-ip-address',
    'source-nat',
    'source-port',
    'source-prefix-list',
    'source-protocol',
    'source-quench',
    'source-route',
    'span',
    'spanning-tree',
    'sparse-dense-mode',
    'sparse-mode',
    'sparse-mode-ssm',
    'spd',
    'spe',
    'spectrum',
    'spectrum-load-balancing',
    'spectrum-monitoring',
    'speed',
    'speed-duplex',
    'spf-interval',
    'spf-options',
    'split-tunnel-network-list',
    'split-tunnel-policy',
    'spt-threshold',
    'sqlnet',
    'sqlnet-v2',
    'src-nat',
    'srlg',
    'srlg-cost',
    'srlg-value',
    'srr-queue',
    'ssh',
    'ssh-certificate',
    'ssh-publickey',
    'ssh_keydir',
    'sshportlist',
    'ssid',
    'ssid-enable',
    'ssid-profile',
    'ssl',
    'ssm',
    'stack-mac',
    'stack-mib',
    'stack-unit',
    'stale-route',
    'standard',
    'standby',
    'start-stop',
    'start-time',
    'startup-query-count',
    'startup-query-interval',
    'stat',
    'state',
    'state-refresh',
    'stateful-dot1x',
    'stateful-kerberos',
    'stateful-ntlm',
    'static',
    'static-group',
    'static-nat',
    'station-address',
    'station-port',
    'station-role',
    'statistics',
    'stbc',
    'stcapp',
    'sticky',
    'sticky-arp',
    'stop',
    'stop-only',
    'stop-record',
    'stopbits',
    'storm-control',
    'storm-control-profiles',
    'stp',
    'streaming',
    'street-address',
    'streetaddress',
    'stricthostkeycheck',
    'string',
    'strip',
    'sts-1',
    'stub',
    'sub-option',
    'sub-route-map',
    'subject-name',
    'submgmt',
    'subnet',
    'subnet-broadcast',
    'subnet-mask',
    'subnet-zero',
    'subnets',
    'subscribe-to',
    'subscribe-to-alert-group',
    'subscriber',
    'substat',
    'subtemplate',
    'subtract',
    'summary-address',
    'summary-lsa',
    'summary-only',
    'sun',
    'sun-rpc',
    'sunrpc',
    'super-user-password',
    'superpassword',
    'supplementary-service',
    'supplementary-services',
    'suppress-arp',
    'suppress-fib-pending',
    'suppressed',
    'suspect-rogue-conf-level',
    'svc',
    'svclc',
    'svp',
    'switch',
    'switch-cert',
    'switch-options',
    'switch-priority',
    'switch-profile',
    'switch-type',
    'switchback',
    'switching-mode',
    'switchname',
    'switchover-on-routing-crash',
    'switchport',
    'symmetric',
    'syn',
    'sync',
    'synchronization',
    'synchronous',
    'syscontact',
    'syslocation',
    'syslog',
    'syslogd',
    'sysopt',
    'system',
    'system-init',
    'system-max',
    'system-priority',
    'system-profile',
    'system-services',
    'system-shutdown',
    'systemname',
    'systemowner',
    'table-map',
    'tac_plus',
    'tacacs',
    'tacacs+',
    'tacacs-ds',
    'tacacs-server',
    'tacplus-server',
    'tacplusprimacctsecret',
    'tacplusprimaddr',
    'tacplusprimauthorsecret',
    'tacplusprimsecret',
    'tacplussecacctsecret',
    'tacplussecaddr',
    'tacplussecauthorsecret',
    'tacplussecsecret',
    'tacplususesub',
    'tag',
    'tag-switching',
    'tag-type',
    'tagged',
    'talk',
    'tap',
    'target',
    'target-host',
    'target-host-port',
    'targeted-broadcast',
    'targets',
    'task',
    'task execute',
    'task-scheduler',
    'taskgroup',
    'tb-vlan1',
    'tb-vlan2',
    'tcam',
    'tcp',
    'tcp-connect',
    'tcp-established',
    'tcp-flags',
    'tcp-initial',
    'tcp-inspection',
    'tcp-mss',
    'tcp-proxy-reassembly',
    'tcp-rst',
    'tcp-session',
    'tcp-udp',
    'tcs-load-balance',
    'te-metric',
    'teardown',
    'telnet',
    'telnet-server',
    'telnetclient',
    'template',
    'term',
    'terminal',
    'terminal-type',
    'termination',
    'test',
    'text',
    'tftp',
    'tftp-server',
    'tftp-server-list',
    'then',
    'threat-detection',
    'threat-visibility',
    'threshold',
    'through',
    'thu',
    'time',
    'time-exceeded',
    'time-format',
    'time-out',
    'time-range',
    'time-zone',
    'timed',
    'timeout',
    'timeouts',
    'timer',
    'timers',
    'timesource',
    'timestamp',
    'timestamp-reply',
    'timestamp-request',
    'timing',
    'tls-proxy',
    'tm-voq-collection',
    'to',
    'to-zone',
    'token',
    'tolerance',
    'tool',
    'top',
    'tos',
    'tos-overwrite',
    'trace',
    'traceoptions',
    'traceroute',
    'track',
    'tracked',
    'tracking-priority-increment',
    'traditional',
    'traffic-eng',
    'traffic-engineering',
    'traffic-index',
    'transceiver',
    'transceiver-type-check',
    'transfer-system',
    'transform-set',
    'translate',
    'translation',
    'translation-profile',
    'translation-rule',
    'transmit',
    'transmit-delay',
    'transparent-hw-flooding',
    'transport',
    'transport-method',
    'transport-mode',
    'trap-destinations',
    'trap-group',
    'trap-options',
    'trap-source',
    'trap-timeout',
    'traps',
    'trigger',
    'trimode',
    'true',
    'trunk',
    'trunk-threshold',
    'trust',
    'trusted',
    'trusted-key',
    'trustpoint',
    'trustpool',
    'tsid',
    'tsm-req-profile',
    'ttl',
    'ttl-exceeded',
    'ttl-threshold',
    'tty',
    'tue',
    'tunable-optic',
    'tunnel',
    'tunnel-group',
    'tunnel-group-list',
    'tunnel-id',
    'tunneled-node-address',
    'tx-queue',
    'txspeed',
    'type',
    'type-1',
    'type-2',
    'type-7',
    'uauth',
    'uc-tx-queue',
    'udf',
    'udld',
    'udp',
    'udp-jitter',
    'udplite',
    'uid',
    'unable',
    'unauthorized-device-profile',
    'unicast',
    'unicast-routing',
    'unique',
    'unit',
    'unnumbered',
    'unreachable',
    'unreachables',
    'untagged',
    'untrust',
    'untrust-screen',
    'update',
    'update-calendar',
    'update-source',
    'upgrade',
    'upgrade-profile',
    'uplink',
    'uplinkfast',
    'upstream',
    'upstream-start-threshold',
    'upto',
    'urg',
    'uri',
    'url',
    'url-list',
    'urpf',
    'urpf-logging',
    'use',
    'use-acl',
    'use-bia',
    'use-ipv4-acl',
    'use-ipv6-acl',
    'use-link-address',
    'use-vrf',
    'user',
    'user-identity',
    'user-message',
    'user-role',
    'user-statistics',
    'usergroup',
    'userinfo',
    'username',
    'username-prompt',
    'userpassphrase',
    'users',
    'using',
    'util-interval',
    'uucp',
    'uuid',
    'v1-only',
    'v1-rp-reachability',
    'v2',
    'v4',
    'v6',
    'vacant-message',
    'vacl',
    'vad',
    'valid-11a-40mhz-channel-pair',
    'valid-11a-80mhz-channel-group',
    'valid-11a-channel',
    'valid-11g-40mhz-channel-pair',
    'valid-11g-channel',
    'valid-and-protected-ssid',
    'valid-network-oui-profile',
    'validation-usage',
    'value',
    'vap-enable',
    'vdc',
    'ver',
    'verify',
    'verify-data',
    'version',
    'vfi',
    'via',
    'video',
    'view',
    'violate-action',
    'violation',
    'virtual',
    'virtual-address',
    'virtual-ap',
    'virtual-chassis',
    'virtual-reassembly',
    'virtual-router',
    'virtual-service',
    'virtual-switch',
    'virtual-template',
    'vlan',
    'vlan-group',
    'vlan-id',
    'vlan-id-list',
    'vlan-name',
    'vlan-policy',
    'vlan-tagging',
    'vlan-tags',
    'vlans',
    'vlt',
    'vlt-peer-lag',
    'vm-cpu',
    'vm-memory',
    'vmps',
    'vmtracer',
    'vocera',
    'voice',
    'voice-card',
    'voice-class',
    'voice-port',
    'voice-service',
    'voip',
    'voip-cac-profile',
    'vpc',
    'vpdn',
    'vpdn-group',
    'vpls',
    'vpn',
    'vpn-dialer',
    'vpn-filter',
    'vpn-group-policy',
    'vpn-idle-timeout',
    'vpn-monitor',
    'vpn-session-timeout',
    'vpn-simultaneous-logins',
    'vpn-tunnel-protocol',
    'vpnv4',
    'vpnv6',
    'vrf',
    'vrf-also',
    'vrf-export',
    'vrf-import',
    'vrf-table-label',
    'vrf-target',
    'vrrp',
    'vrrp-group',
    'vserver',
    'vstack',
    'vstp',
    'vti',
    'vtp',
    'vty',
    'vty-pool',
    'vxlan',
    'wait-for',
    'wait-start',
    'warnings',
    'warntime',
    'watch-list',
    'watchdog',
    'wavelength',
    'wccp',
    'web-cache',
    'web-https-port-443',
    'web-max-clients',
    'web-server',
    'webvpn',
    'wed',
    'weekday',
    'weekend',
    'weight',
    'weighting',
    'welcome-page',
    'white-list',
    'who',
    'whois',
    'wide',
    'wide-metric',
    'wide-metrics-only',
    'wideband',
    'wildcard-address',
    'window-size',
    'wins-server',
    'wired',
    'wired-ap-profile',
    'wired-containment',
    'wired-port-profile',
    'wired-to-wireless-roam',
    'wireless',
    'wireless-containment',
    'wirelessmodem',
    'wism',
    'wispr',
    'without-csd',
    'wlan',
    'wmm',
    'wms-general-profile',
    'wms-local-system-profile',
    'wpa-fast-handover',
    'wred',
    'wred-profile',
    'write-memory',
    'wrr',
    'wrr-queue',
    'wsma',
    'www',
    'x25',
    'x29',
    'x509',
    'xauth',
    'xconnect',
    'xdmcp',
    'xdr',
    'xlate',
    'xml-config',
    'xnm-clear-text',
    'xnm-ssl',
    'yellow',
    'zone',
    'zones'
}



"""Generate & apply default regexes for finding & removing sensitive info."""
#   Copyright 2018 Intentionet
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from __future__ import absolute_import
# Need regex here instead of re for variable length lookbehinds
import regex
import logging

from binascii import b2a_hex
from enum import Enum
from hashlib import md5
from .default_pwd_regexes import default_pwd_line_regexes, default_com_line_regexes
from .default_reserved_words import default_reserved_words
# Using passlib for digests not supported by hashlib
from passlib.hash import cisco_type7, md5_crypt, sha512_crypt
from six import b


# A regex matching any of the characters that are allowed to precede a password
# regex (e.g. sensitive line is allowed to be in quotes or after a colon)
# This is an ignored group, so it does not muck with the password regex indicies
_ALLOWED_REGEX_PREFIX = r'(?:[^-_a-zA-Z\d] ?|^ ?)'

# Number of digits to extract from hash for sensitive keyword replacement
_ANON_SENSITIVE_WORD_LEN = 6

# BGP communities should be ignored/not anonymized, and can be detected by the
# following patterns, mostly extracted from examples at
# https://www.cisco.com/c/en/us/td/docs/routers/crs/software/crs_r4-1/routing/command/reference/b_routing_cr41crs/b_routing_cr41crs_chapter_01000.html
# Text followed by the word 'additive'
_IGNORED_COMM_ADDITIVE = r'\S+ additive'
# Numeric, colon separated, and parameter ($) communities
_IGNORED_COMM_COLON = r'(peeras|\$\w+|\d+)\:(peeras|\$\w+|\d+)'
# List of communities enclosed in parenthesis, being permissive here for the
# content inside the parenthesis for simplicity
_IGNORED_COMM_LIST = r'\([\S ]+\)'
# Well-known BPG communities
_IGNORED_COMM_WELL_KNOWN = 'gshut|internet|local-AS|no-advertise|no-export|none'
_IGNORED_COMMUNITIES = (r'((\d+|{additive}|{colon}|{list}|{well_known})(?!\S))'
                        .format(additive=_IGNORED_COMM_ADDITIVE,
                                colon=_IGNORED_COMM_COLON,
                                list=_IGNORED_COMM_LIST,
                                well_known=_IGNORED_COMM_WELL_KNOWN))

# Text that is allowed to surround passwords, to be preserved
_PASSWORD_ENCLOSING_TEXT = ['\'', '"', '\\\'', '\\"']

# These are extra regexes to find lines that seem like they might contain
# sensitive info (these are not already caught by RANCID default regexes)
extra_password_regexes = [
    [(r'encrypted-password \K(\S+)', None)],
    [(r'key "\K([^"]+)', 1)],
    [(r'key-hash sha256 (\S+)', 1)],
    # Replace communities that do not look like well-known BGP communities
    # i.e. snmp communities
    [(r'set community \K((?!{ignore})\S+)'
      .format(ignore=_IGNORED_COMMUNITIES), 1)],
    [(r'snmp-server mib community-map \K([^ :]+)', 1)],
    [(r'snmp-community \K(\S+)', 1)],
    # Catch-all's matching what looks like hashed passwords
    [(r'\K("?\$9\$[^\s;"]+)', 1)],
    [(r'\K("?\$1\$[^\s;"]+)', 1)],
]


class AsNumberAnonymizer(object):
    """An anonymizer for AS numbers."""

    # AS number block boundaries - each number corresponds to beginning of the next AS num block
    # Except the last, which just serves to indicate the end of the previous block
    _AS_NUM_BOUNDARIES = [0, 64512, 65536, 4200000000, 4294967296]

    def __init__(self, as_numbers, salt):
        """Create an anonymizer for the specified list of AS numbers (strings) and salt."""
        self.salt = salt
        self._generate_as_number_regex(as_numbers)
        self._generate_as_number_replacement_map(as_numbers)

    def anonymize(self, as_number):
        """Anonymize the specified AS number (string)."""
        return self.as_num_map[as_number]

    def _generate_as_number_regex(self, as_numbers):
        """Generate regex for finding AS number."""
        # Match a non-digit, any of the AS numbers and another non-digit
        # Using lookahead and lookbehind to match on context but not include that context in the match
        self.as_num_regex = regex.compile(r'(\D|^)\K({})(?=\D|$)'.format(
            '|'.join(as_numbers)))

    def _generate_as_number_replacement(self, as_number):
        """Generate a replacement AS number for the given AS number and salt."""
        hash_val = int(md5((self.salt + as_number).encode()).hexdigest(), 16)
        as_number = int(as_number)
        if as_number < 0 or as_number > 4294967295:
            raise ValueError('AS number provided was outside accepted range (0-4294967295)')

        block_begin = 0
        for next_block_begin in self._AS_NUM_BOUNDARIES:
            if as_number < next_block_begin:
                return str(hash_val % (next_block_begin - block_begin) + block_begin)
            block_begin = next_block_begin

    def _generate_as_number_replacement_map(self, as_numbers):
        """Generate map of AS numbers and their replacements."""
        self.as_num_map = {as_num: self._generate_as_number_replacement(as_num) for as_num in as_numbers}

    def get_as_number_pattern(self):
        """Return the compiled regex to find AS numbers."""
        return self.as_num_regex


class SensitiveWordAnonymizer(object):
    """An anonymizer for sensitive keywords."""

    def __init__(self, sensitive_words, salt, reserved_words=default_reserved_words):
        """Create an anonymizer for specified list of sensitive words and set of reserved words to leave alone."""
        self.reserved_words = reserved_words
        self.sens_regex = self._generate_sensitive_word_regex(sensitive_words)
        self.sens_word_replacements = self._generate_sensitive_word_replacements(sensitive_words, salt)
        # Figure out which reserved words may clash with sensitive words, so they can be preserved in anonymization
        self.conflicting_words = self._generate_conflicting_reserved_word_list(sensitive_words)

    def anonymize(self, line):
        """Anonymize sensitive words from the input line."""
        if self.sens_regex.search(line) is not None:
            leading, words, trailing = _split_line(line)
            # Anonymize only words that do not match the conflicting (reserved) words
            words = [
                w if w in self.conflicting_words else self.sens_regex.sub(self._lookup_anon_word, w) for w in words
            ]
            # Restore leading and trailing whitespace since those were removed when splitting into words
            line = leading + ' '.join(words) + trailing
        return line

    def _generate_conflicting_reserved_word_list(self, sensitive_words):
        """Return a list of reserved words that may conflict with the specified sensitive words."""
        conflicting_words = set()
        for sensitive_word in sensitive_words:
            conflicting_words.update(set([w for w in self.reserved_words if sensitive_word in w]))
        if conflicting_words:
            logging.warning('Specified sensitive words overlap with reserved words. '
                            'The following reserved words will be preserved: %s', conflicting_words)
        return conflicting_words

    @classmethod
    def _generate_sensitive_word_regex(cls, sensitive_words):
        """Compile and return regex for the specified list of sensitive words."""
        return regex.compile('({})'.format('|'.join(sensitive_words)), regex.IGNORECASE)

    @classmethod
    def _generate_sensitive_word_replacements(cls, sensitive_words, salt):
        """Compile and return a dict of sensitive word replacements."""
        # Only using part of the md5 hash result as the anonymized replacement
        # to cut down on the size of the replacements
        return {
            sens_word.lower(): md5((salt + sens_word.lower()).encode()).hexdigest()[:_ANON_SENSITIVE_WORD_LEN]
            for sens_word in sensitive_words
        }

    def _lookup_anon_word(self, match):
        """Lookup anonymized word for the given sensitive word regex match."""
        return self.sens_word_replacements[match.group(0).lower()]


class _sensitive_item_formats(Enum):
    """Enum for recognized sensitive item formats (e.g. type7, md5, text)."""

    cisco_type7 = 1
    numeric = 2
    hexadecimal = 3
    md5 = 4
    text = 5
    sha512 = 6
    juniper_type9 = 7


def anonymize_as_numbers(anonymizer, line):
    """Anonymize AS numbers in the input line."""
    as_number_regex = anonymizer.get_as_number_pattern()
    return as_number_regex.sub(lambda match: anonymizer.anonymize(match.group(0)), line)


def anonymize_sensitive_words(sensitive_word_regexes, line, salt):
    """Anonymize words from specified sensitive words list in the input line."""
    for sens_word_regex in sensitive_word_regexes:
        if sens_word_regex.search(line) is not None:
            sens_word = sens_word_regex.pattern
            # Only using part of the hash result as the anonymized replacement
            # to cut down on the size of the replacements
            anon_word = md5((salt + sens_word).encode()).hexdigest()[:_ANON_SENSITIVE_WORD_LEN]
            line = sens_word_regex.sub(anon_word, line)
    return line


def _anonymize_value(val, lookup):
    """Generate an anonymized replacement for the input value.

    This function tries to determine what type of value was passed in and
    returns an anonymized value of the same format.  If the source value has
    already been anonymized in the provided lookup, then the previous anon
    value will be used.
    """
    # Separate enclosing text (e.g. quotes) from the underlying value
    enclosing_text, val = _extract_enclosing_text(val)
    item_format = _check_sensitive_item_format(val)

    anon_val = 'netconanRemoved{}'.format(len(lookup))
    if val in lookup:
        return enclosing_text + lookup[val] + enclosing_text

    if item_format == _sensitive_item_formats.cisco_type7:
        # Not salting sensitive data, using static salt here to more easily
        # identify anonymized lines
        anon_val = cisco_type7.using(salt=9).hash(anon_val)

    if item_format == _sensitive_item_formats.numeric:
        # These are the ASCII character values for anon_val converted to decimal
        anon_val = str(int(b2a_hex(b(anon_val)), 16))

    if item_format == _sensitive_item_formats.hexadecimal:
        # These are the ASCII character values for anon_val in hexadecimal
        anon_val = b2a_hex(b(anon_val)).decode()

    if item_format == _sensitive_item_formats.md5:
        old_salt_size = len(val.split('$')[2])
        # Not salting sensitive data, using static salt here to more easily
        # identify anonymized lines
        anon_val = md5_crypt.using(salt='0' * old_salt_size).hash(anon_val)

    if item_format == _sensitive_item_formats.sha512:
        # Hash anon_val w/standard rounds=5000 to omit rounds parameter from hash output
        anon_val = sha512_crypt.using(rounds=5000).hash(anon_val)

    if item_format == _sensitive_item_formats.juniper_type9:
        # TODO(https://github.com/intentionet/netconan/issues/16)
        # Encode base anon_val instead of just returning a constant here
        # This value corresponds to encoding: Conan812183
        anon_val = '$9$0000IRc-dsJGirewg4JDj9At0RhSreK8Xhc'

    lookup[val] = anon_val
    return enclosing_text + anon_val + enclosing_text


def _check_sensitive_item_format(val):
    """Determine the type/format of the value passed in."""
    item_format = _sensitive_item_formats.text

    # Order is important here (e.g. type 7 looks like hex or text, but has a
    # specific format so it should override hex or text)
    if regex.match(r'^\$9\$[\S]+$', val):
        item_format = _sensitive_item_formats.juniper_type9
    if regex.match(r'^\$6\$[\S]+$', val):
        item_format = _sensitive_item_formats.sha512
    if regex.match(r'^\$1\$[\S]+\$[\S]+$', val):
        item_format = _sensitive_item_formats.md5
    if regex.match(r'^[0-9a-fA-F]+$', val):
        item_format = _sensitive_item_formats.hexadecimal
    if regex.match(r'^[01][0-9]([0-9a-fA-F]{2})+$', val):
        item_format = _sensitive_item_formats.cisco_type7
    if regex.match(r'^[0-9]+$', val):
        item_format = _sensitive_item_formats.numeric
    return item_format


def _extract_enclosing_text(val):
    """Extract enclosing quotes from text and return the enclosing text and enclosed text."""
    enclosing_text = ''
    for surround_text in _PASSWORD_ENCLOSING_TEXT:
        if val.endswith(surround_text) and val.startswith(surround_text):
            enclosing_text = surround_text
            val = val[len(surround_text):-len(surround_text)]
            break
    return enclosing_text, val


def generate_default_sensitive_item_regexes():
    """Compile and return the default password and community line regexes."""
    combined_regexes = default_pwd_line_regexes + default_com_line_regexes + \
        extra_password_regexes
    return [[(regex.compile(_ALLOWED_REGEX_PREFIX + regex_), num) for regex_, num in group]
            for group in combined_regexes]


def replace_matching_item(compiled_regexes, input_line, pwd_lookup, reserved_words=default_reserved_words):
    """If line matches a regex, anonymize or remove the line."""
    # Collapse whitespace to simplify regexes, also preserve leading and trailing whitespace
    leading, words, trailing = _split_line(input_line)
    output_line = ' '.join(words)

    # Note: compiled_regexes is a list of lists; the inner list is a group of
    # related regexes
    for compiled_regex_grp in compiled_regexes:
        match_found = False

        # Apply all related regexes before returning the output_line
        for compiled_re, sensitive_item_num in compiled_regex_grp:
            # Using search instead of match here to find the match anywhere in the input line
            match = compiled_re.search(output_line)
            if match is None:
                continue
            match_found = True
            logging.debug('Match found on %s', output_line.rstrip())

            # If this regex cannot preserve text around sensitive info,
            # then just remove the whole line
            if sensitive_item_num is None:
                logging.warning(
                    'Anonymizing sensitive info in lines like "%s" is currently'
                    ' unsupported, so removing this line completely',
                    compiled_re.pattern)
                return '! Sensitive line SCRUBBED by netconan\n'

            sensitive_val = match.group(sensitive_item_num)
            if sensitive_val in reserved_words:
                logging.debug('Skipping anonymization of reserved word: "%s"', sensitive_val)
                continue
            anon_val = _anonymize_value(sensitive_val, pwd_lookup)
            output_line = compiled_re.sub(anon_val, output_line)
            logging.debug(
                'Anonymized input "%s" to "%s"', sensitive_val, anon_val)

        # If any matches existed in this regex group, stop processing more regexes
        if match_found:
            break

    # Restore leading and trailing whitespace for readability and context
    return leading + output_line + trailing


def _split_line(line):
    """Split line into leading whitespace, list of words, and trailing whitespace."""
    return line[:-len(line.lstrip())], line.split(), line[len(line.rstrip()):]



"""Default regexes from RANCID for finding & removing passwords."""
# Copyright (c) 1997-2017 by Henry Kilmer and John Heasley
# All rights reserved.
#
# This code is derived from software contributed to and maintained by
# Henry Kilmer, John Heasley, Andrew Partan,
# Pete Whiting, Austin Schutz, and Andrew Fort.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. Neither the name of RANCID nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY Henry Kilmer, John Heasley AND CONTRIBUTORS
# ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COMPANY OR CONTRIBUTORS
# BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.


# Regexes taken from RANCID password scrubbing
#
# Multiple regexes in a single inner list will all be run before returning
# the anonymized line.  This is useful for config lines that may have multiple
# sensitive items (e.g. snmp-server auth password and priv password).
#
# Format for these tuples are:
#  1. sensitive line regex
#       note that the regexes use lookbehinds and lookaheads so we can easily
#       extract and replace just the sensitive information
#  2. sensitive item regex-match-index
#       note that if this is None, any matching config line will be removed
default_pwd_line_regexes = [
    [(r'((password|passwd)( level \d+)?( \d+)?) \K(\S+)', 5)],
    [(r'(username( \S+)+ (password|secret)( \d| sha512)?) \K(\S+)', 5)],
    [(r'((enable )?secret( \d)?) \K(\S+)', 4)],
    [(r'(ip ftp password( \d)?) \K(\S+)', 3)],
    [(r'(ip ospf authentication-key( \d)?) \K(\S+)', 3)],
    [(r'(isis password) \K(\S+)(?=( level-\d)?)', 2)],
    [(r'((domain-password|area-password)) \K(\S+)', 3)],
    [(r'(ip ospf message-digest-key \d+ md5( \d)?) \K(\S+)', 3)],
    [(r'(standby( \d*)? authentication( text| md5 key-string( \d)?)?) \K(\S+)', 5)],
    [(r'(l2tp tunnel( \S+)? password( \d)?) \K(\S+)', 4)],
    [(r'(digest secret( \d)?) \K(\S+)', 3)],
    [(r'(ppp .* hostname) \K(\S+)', 2)],
    [(r'(ppp .* password( \d)?) \K(\S+)', 3)],
    [(r'((ikev2 )?(local|remote)-authentication pre-shared-key) \K(\S+)', 4)],
    [(r'((\S )*pre-shared-key( remote| local)?( hex| hexadecimal| ascii-text| \d)?) \K(\S+)', 5)],
    [(r'((tacacs|radius)-server (\S+ )*key)( \d)? \K(\S+)', 5)],
    [(r'(key( \d)?) \K(\S+)', 3)],
    [(r'(ntp authentication-key \d+ md5) \K(\S+)', 2)],
    [(r'(syscon( password| address \S+)) \K(\S+)', 3)],
    [(r'(snmp-server user( \S+)+ (auth (md5|sha))) \K(\S+)', 5),
     (r'(snmp-server user( \S+)+ priv( 3des| aes( \d+)?| des)?) \K(\S+)', 5)],
    [(r'((crypto )?isakmp key( \d)?) \K(\S+)', 4)],
    [(r'(set session-key (in|out)bound ah \d+) \K(\S+)', 3)],
    [(r'(set session-key (in|out)bound esp \d+ cipher?) \K(\S+)', 3),
     (r'(set session-key (in|out)bound esp \d+(( cipher \S+)? authenticator)) \K(\S+)', 5)],
    # TODO(https://github.com/intentionet/netconan/issues/3):
    # Follow-up on these.  They were just copied from RANCID so currently:
    #   They are untested in general and need cases added for unit tests
    #   They do not specifically capture sensitive info
    #   They just identify lines where sensitive info exists
    [(r'(cable shared-secret) (.*)', None)],
    [(r'(wpa-psk ascii|hex \d) (.*)', None)],
    [(r'(ldap-login-password) \S+(.*)', None)],
    [(r'((ikev1 )?(pre-shared-key |key |failover key )(ascii-text |hexadecimal )?).*(.*)', None)],
    [(r'(vpdn username (\S+) password)(.*)', None)],
    [(r'(key-string \d?)(.*)', None)],
    [(r'(message-digest-key \d+ md5 (7|encrypted)) (.*)', None)],
    [(r'(.*?neighbor.*?) (\S*) password (.*)', None)],
    [(r'(wlccp \S+ username (\S+)( .*)? password( \d)?) (\S+)(.*)', None)],

    # These are regexes for JUNOS
    # TODO(https://github.com/intentionet/netconan/issues/4):
    # Follow-up on these.  They were modified from RANCID's regexes and currently:
    #   They do not have capture groups for sensitive info
    #   They just identify lines where sensitive info exists
    #   They need to be tested against config lines generated on a JUNOS router
    #     (to make sure the regex handles different syntaxes allowed in the line)
    [(r'(\S* )*authentication-key [^ ;]+(.*)', None)],
    [(r'(\S* )*md5 \d+ key [^ ;]+(.*)', None)],
    [(r'(\S* )*hello-authentication-key [^ ;]+(.*)', None)],
    [(r'(\S* )*(secret|simple-password) [^ ;]+(.*)', None)],
    [(r'(\S* )*encrypted-password [^ ;]+(.*)', None)],
    [(r'(\S* )*ssh-(rsa|dsa) \"(.*)', None)],
    [(r'(\S* )*((pre-shared-|)key (ascii-text|hexadecimal)) [^ ;]+(.*)', None)]
]
# Taken from RANCID community scrubbing regexes
default_com_line_regexes = [
    [(r'((snmp-server (\S+ )*community)( [08])?) \K(\S+)', 5)],
    # TODO(https://github.com/intentionet/netconan/issues/5):
    # Confirm this catches all community possibilities for snmp-server
    [(r'(snmp-server host (\S+)( informs| traps| version '
     r'(?:1|2c|3 \S+)| vrf \S+)*) \K(\S+)', 4)],
    # This is from JUNOS
    # TODO(https://github.com/intentionet/netconan/issues/4):
    # See if we need to make the snmp keyword optional for Juniper
    # Also, this needs to be tested against config lines generated on a JUNOS router
    #     (to make sure the regex handles different syntaxes allowed in the line)
    [(r'((\S* )*snmp( \S+)* (community|trap-group)) \K([^ ;]+)', 5)]
]



"""A setuptools based setup module.

See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""
#   Copyright 2018 Intentionet
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
from os import path

here = path.abspath(path.dirname(__file__))

about = {}
with open(path.join(here, 'netconan', '__version__.py'), 'r') as f:
    exec(f.read(), about)

with open(path.join(here, 'README.rst')) as f:
    readme = f.read()

setup(
    name=about['__name__'],
    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version=about['__version__'],
    description=about['__desc__'],
    long_description=readme,
    # The project's main homepage.
    url=about['__url__'],
    # Author details
    author='Intentionet',
    author_email='netconan-dev@intentionet.com',

    # Choose your license
    license='Apache License 2.0',

    # What does your project relate to?
    keywords='network configuration anonymizer',

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=find_packages(exclude=['contrib', 'docs', 'tests']),

    # Alternatively, if you want to distribute just a my_module.py, uncomment
    # this:
    #   py_modules=["my_module"],

    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=[
        'configargparse', 'bidict', 'enum34', 'ipaddress', 'passlib', 'regex', 'six'
    ],

    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e .[dev,test]
    extras_require={
        'dev': ['flake8', 'flake8-docstrings'],
        # 'test': ['coverage', 'pytz'],
    },

    # List pytest requirements for running unit tests
    setup_requires=['pytest-runner'],
    tests_require=['pytest', 'pytest-cov', 'requests_mock', 'testfixtures'],

    # If there are data files included in your packages that need to be
    # installed, specify them here.  If using Python 2.6 or less, then these
    # have to be included in MANIFEST.in as well.
    package_data={},

    # Although 'package_data' is the preferred approach, in some case you may
    # need to place data files outside of your packages. See:
    # http://docs.python.org/3.4/distutils/setupscript.html#installing-additional-files # noqa
    # In this case, 'data_file' will be installed into '<sys.prefix>/my_data'
    data_files=[],

    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # pip to create the appropriate form of executable for the target platform.
    entry_points={
        'console_scripts': [
            'netconan = netconan.netconan:main',
        ],
    },
)



"""Anonymize configuration file(s)."""
#   Copyright 2018 Intentionet
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from __future__ import absolute_import
import errno
import logging
import os
import random
import string

from .default_reserved_words import default_reserved_words
from .ip_anonymization import (
    IpAnonymizer, IpV6Anonymizer, anonymize_ip_addr)
from .sensitive_item_removal import (
    anonymize_as_numbers, AsNumberAnonymizer, replace_matching_item,
    SensitiveWordAnonymizer, generate_default_sensitive_item_regexes)

_DEFAULT_SALT_LENGTH = 16
_CHAR_CHOICES = string.ascii_letters + string.digits


def anonymize_files(input_path, output_path, anon_pwd, anon_ip,
                    salt=None, dumpfile=None, sensitive_words=None,
                    undo_ip_anon=False, as_numbers=None, reserved_words=None):
    """Anonymize each file in input and save to output."""
    anonymizer4 = None
    anonymizer6 = None
    anonymizer_as_num = None
    anonymizer_sensitive_word = None
    compiled_regexes = None
    pwd_lookup = None
    # The salt is only used for IP and sensitive word anonymization:
    if salt is None:
        salt = ''.join(random.choice(_CHAR_CHOICES) for _ in range(_DEFAULT_SALT_LENGTH))
        logging.warning('No salt was provided; using randomly generated "%s"', salt)
    logging.debug('Using salt: "%s"', salt)
    if anon_pwd:
        compiled_regexes = generate_default_sensitive_item_regexes()
        pwd_lookup = {}
    if reserved_words is not None:
        default_reserved_words.update(reserved_words)

    if sensitive_words is not None:
        anonymizer_sensitive_word = SensitiveWordAnonymizer(sensitive_words, salt)
    if anon_ip or undo_ip_anon:
        anonymizer4 = IpAnonymizer(salt)
        anonymizer6 = IpV6Anonymizer(salt)
    if as_numbers is not None:
        anonymizer_as_num = AsNumberAnonymizer(as_numbers, salt)

    if not os.path.exists(input_path):
        raise ValueError("Input does not exist")

    # Generate list of file tuples: (input file path, output file path)
    file_list = []
    if os.path.isfile(input_path):
        file_list = [(input_path, output_path)]
    else:
        if not os.listdir(input_path):
            raise ValueError("Input directory is empty")
        if os.path.isfile(output_path):
            raise ValueError("Output path must be a directory if input path is "
                             "a directory")

        for root, dirs, files in os.walk(input_path):
            rel_root = os.path.relpath(root, input_path)
            file_list.extend([(
                os.path.join(input_path, rel_root, f),
                os.path.join(output_path, rel_root, f)
            ) for f in files if not f.startswith('.')])

    for in_path, out_path in file_list:
        try:
            anonymize_file(in_path,
                           out_path,
                           compiled_regexes=compiled_regexes,
                           pwd_lookup=pwd_lookup,
                           anonymizer_sensitive_word=anonymizer_sensitive_word,
                           anonymizer_as_num=anonymizer_as_num,
                           undo_ip_anon=undo_ip_anon,
                           anonymizer4=anonymizer4,
                           anonymizer6=anonymizer6)
        except Exception:
            logging.error('Failed to anonymize file %s', in_path, exc_info=True)

    if dumpfile is not None:
        with open(dumpfile, 'w') as f_out:
            anonymizer4.dump_to_file(f_out)
            anonymizer6.dump_to_file(f_out)


def anonymize_file(filename_in, filename_out, compiled_regexes=None,
                   anonymizer4=None, anonymizer6=None, pwd_lookup=None,
                   anonymizer_sensitive_word=None, anonymizer_as_num=None,
                   undo_ip_anon=False):
    """Anonymize contents of input file and save to the output file.

    This only applies sensitive line removal if compiled_regexes and pwd_lookup
    are not None.  This only applies ip anonymization if anonymizer is not None.
    """
    logging.debug("File in  %s", filename_in)
    logging.debug("File out %s", filename_out)

    # Make parent dirs for output file if they don't exist
    _mkdirs(filename_out)

    if os.path.isdir(filename_out):
        raise ValueError('Cannot write output file; '
                         'output file is a directory ({})'
                         .format(filename_out))

    with open(filename_out, 'w') as f_out, open(filename_in, 'r') as f_in:
        for line in f_in:
            output_line = line
            if compiled_regexes is not None and pwd_lookup is not None:
                output_line = replace_matching_item(compiled_regexes,
                                                    output_line, pwd_lookup)

            if anonymizer6 is not None:
                output_line = anonymize_ip_addr(anonymizer6, output_line, undo_ip_anon)
            if anonymizer4 is not None:
                output_line = anonymize_ip_addr(anonymizer4, output_line, undo_ip_anon)

            if anonymizer_sensitive_word is not None:
                output_line = anonymizer_sensitive_word.anonymize(output_line)

            if anonymizer_as_num is not None:
                output_line = anonymize_as_numbers(anonymizer_as_num, output_line)

            if line != output_line:
                logging.debug("Input line:  %s", line.rstrip())
                logging.debug("Output line: %s", output_line.rstrip())
            f_out.write(output_line)


def _mkdirs(file_path):
    """Make parent directories for the specified file if they don't exist."""
    dir_path = os.path.dirname(file_path)
    if len(dir_path) > 0:
        try:
            os.makedirs(dir_path)
        except OSError as e:
            if e.errno == errno.EEXIST and os.path.isdir(dir_path):
                pass
            else:
                raise



"""Handle invoking netconan from the command line."""
#   Copyright 2018 Intentionet
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from __future__ import absolute_import
import configargparse
import logging
import sys

from .anonymize_files import anonymize_files


def _parse_args(argv):
    """Parse arguments from the given list."""
    parser = configargparse.ArgParser(
        # Replace the default config file help with custom message
        # To fix some syntax issues
        add_config_file_help=False,
        description="""
        Args that can start with '--' can also be set in a config file (specified
        via -c). If an arg is specified in more than one place, then command line
        values override config file values which override defaults. Config file
        syntax allows: key=value, flag=true, stuff=[a,b,c] (for more details, see
        here https://goo.gl/R74nmi).
        """
    )

    parser.add_argument('-a', '--anonymize-ips', action='store_true', default=False,
                        help='Anonymize IP addresses')
    parser.add_argument('-c', '--config', is_config_file=True,
                        help='Config file specifying params')
    parser.add_argument('-d', '--dump-ip-map', default=None,
                        help='Dump IP address anonymization map to specified file')
    parser.add_argument('-i', '--input', required=True,
                        help='Input file or directory containing files to anonymize')
    parser.add_argument('-l', '--log-level', default='INFO',
                        choices=['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'],
                        help='Determines what level of logs to display')
    parser.add_argument('-n', '--as-numbers', default=None,
                        help='List of comma separated AS numbers to anonymize')
    parser.add_argument('-o', '--output', required=True,
                        help='Output file or directory where anonymized files are placed')
    parser.add_argument('-p', '--anonymize-passwords', action='store_true', default=False,
                        help='Anonymize password and snmp community lines')
    parser.add_argument('-r', '--reserved-words', default=None,
                        help='List of comma separated words that should not be anonymized')
    parser.add_argument('-s', '--salt', default=None,
                        help='Salt for IP and sensitive keyword anonymization')
    parser.add_argument('-u', '--undo', action='store_true', default=False,
                        help='Undo reversible anonymization (must specify salt)')
    parser.add_argument('-w', '--sensitive-words', default=None,
                        help='List of comma separated keywords to anonymize')
    return parser.parse_args(argv)


def main(argv=sys.argv[1:]):
    """Netconan tool entry point."""
    args = _parse_args(argv)

    if not args.input:
        raise ValueError("Input must be specified")

    log_level = logging.getLevelName(args.log_level)
    logging.basicConfig(format='%(levelname)s %(message)s', level=log_level)

    if not args.output:
        raise ValueError("Output must be specified")

    if args.undo:
        if args.anonymize_ips:
            raise ValueError('Cannot anonymize and undo anonymization, select '
                             'only one.')
        if args.salt is None:
            raise ValueError('Salt used for anonymization must be specified in '
                             'order to undo anonymization.')

    if args.dump_ip_map is not None:
        if not args.anonymize_ips:
            raise ValueError('Can only dump IP address map when anonymizing IP '
                             'addresses.')

    as_numbers = None
    if args.as_numbers is not None:
        as_numbers = args.as_numbers.split(',')

    reserved_words = None
    if args.reserved_words is not None:
        reserved_words = args.reserved_words.split(',')

    sensitive_words = None
    if args.sensitive_words is not None:
        sensitive_words = args.sensitive_words.split(',')

    anonymize_files(args.input, args.output, args.anonymize_passwords, args.anonymize_ips, args.salt,
                    args.dump_ip_map, sensitive_words, args.undo, as_numbers, reserved_words)


if __name__ == '__main__':
    main()
